#!/usr/bin/perl -w
use subs;
use strict;
use File::Path;
use List::Util qw(shuffle);
use FindBin qw($RealDir);
use lib $RealDir;
use MISC qw(:other);
use FORKER;
use TauP;
use SAC;

# HELP MESSAGE
if ($#ARGV!=4) {
	my $help=<<'_END_';

#### DESCRIPTION #######################################################
########################################################################
##   Gets the 1D estimated arrival time.  Puts travel time, phase     ##
##  and ray parameter in fields tN,ktN,userN where N is one of 0-9.   ##
##  Requires TauP!                                                    ##
##                                                                    ##
##   LAST EDITED: Nov. 26, 2007 at  2:45pm by Garrett Euler           ##
##   VERSION INFO:                                                    ##
##            3.2: sac macro sig dig fix                              ##
##            3.1: threading support added to optimize efficiency     ##
##            3.0: TauP perl module does the labor                    ##
##            2.2: error checks added                                 ##
##            2.1: faster for diffracted arrivals                     ##
##            2.0: fixed to work with new geometry format             ##
##            1.1: streamlined and added this helpful feature         ##
##            1.0: initial version                                    ##
########################################################################

#### USAGE #############################################################
########################################################################
##  Command line input:                                               ##
##     arrival.pl in_data_dir new_data_dir phase model nproc          ##
##                                                                    ##
##     where:                                                         ##
##        *in_data_dir - input data directory                         ##
##        *new_data_dir - output data directory                       ##
##        *phase - seismic phase of interest (case sensitive)         ##
##        *model - 1D model for travel time calculation               ##
##        *nproc - number of processing threads for this task         ##
########################################################################

_END_
	print $help;
	exit(1);
}

# DATA DIR, NEW DATA DIR, PHASE, 1D MODEL, NUMBER OF PROCESSES
my ($ddir,$ndir,$phase,$rmod,$nproc)=@ARGV;

# CHECK INPUT DIRECTORY EXISTS
if (! -e $ddir) {print "DATA DIRECTORY $ddir NOT FOUND!\n" and exit(2);}

# CREATE OUTPUT DIRECTORY
unless (defined eval {mkpath($ndir,0,0775)}) {print "COULD NOT CREATE DIRECTORY: $ndir!\n" and exit(3);}

# GETTING ABSOLUTE PATHS TO DIRECTORIES
my ($andir)=abs_path_list($ndir);

# GET DATAFILE LIST
my @files=get_dir_files($ddir);
@files=shuffle(@files);

# MOVING INTO DATA DIRECTORY
chdir($ddir) or print "COULD NOT MOVE INTO DATA DIRECTORY: $ddir!\n" and exit(4);

# GET HEADER INFO
my @data=parselh(saclh(['evdp','gcarc'],[@files]));
my @data2=parselh(saclh(['picks'],[@files]));
my @data3=parselh(saclh(['user0','user1','user2','user3','user4','user5','user6','user7','user8','user9'],[@files]));

# CHECK EVDP IS THE SAME
my ($evflag,$evdp)=(0,${$data[0]}[1]);
foreach (@data) {if ($$_[1]!=$evdp) {print "MULTIPLE EVENT DEPTHS FOUND - VERY SLOW FOR DIFFRACTED ARRIVALS!\n"; $evflag=1;}}

# SPECIAL SECTION FOR CORE-DIFFRACTED ARRIVALS TO SPEED CALCULATIONS
my ($tt,$rp,$dtt,$drp);
if (!$evflag and $phase eq 'Pdiff' or $phase eq 'Sdiff') {
	# RUN TAUP FOR AN AVAILABLE DISTANCE
	($dtt,$drp)=taup_time($rmod,$evdp/1000,130,$phase);
}

# FORK OUT
my $id=forker($nproc);

# ASSIGN PARSELS
@data=parsel($id,$nproc,@data);
@data2=parsel($id,$nproc,@data2);
@data3=parsel($id,$nproc,@data3);

# CHECKING EACH FILE
for (my $i=0; $i<=$#data; $i++) {
	# EXPAND RECORD
	my ($file,$evdp,$gcarc)=@{$data[$i]};
	
	# FIND AN OPEN PICK SLOT
	my $pick;
	for (my $j=1; $j<=10; $j++) {
		if (${$data2[$i]}[$j]==-12345 and ${$data3[$i]}[$j]==-12345) {
			$pick=$j-1; last;
		}
	}
	if (!defined($pick)) {print "NO OPEN PICK FIELDS FOR $file - SKIPPING!\n"; next;}
	
	# CORE-DIFFRACTED?
	if (!$evflag and $phase eq 'Pdiff' or $phase eq 'Sdiff') {
		# CALCULATE TRAVEL TIME
		$tt=$dtt+($gcarc-130)*$drp; $rp=$drp;
	} else {
		# RUN TAUP
		($tt,$rp)=taup_time($rmod,$evdp/1000,$gcarc,$phase);
	}
	
	# SIG DIG FIX
	$tt=sprintf('%12.9g',$tt);
	$rp=sprintf('%12.9g',$rp);
	
	# MAKE MACRO
	(my $sac=<<"	_END_")=~s/^\s+//gm;
		production on
		r $file
		chnhdr t$pick $tt
		chnhdr kt$pick $phase
		chnhdr user$pick $rp
		w dir $andir over
		quit
	_END_

	# RUN MACRO
	runsac($sac);
	
	# VERIFY
	if (! -e $andir.$file) {print "FAILED TO INJECT ARRIVAL DATA INTO $file!\n" and mexit(5);}
}

# MULTI-EXIT
mexit(0);

