#!/usr/bin/perl -w
use subs;
use strict;
use File::Path;
use Time::Local;
use List::Util qw(shuffle);
use FindBin qw($RealDir);
use lib $RealDir;
use MISC qw(:other open_table);
use DATETIME;
use FORKER;

# HELP MESSAGE
if ($#ARGV!=8) {
	my $help=<<'_END_';

#### DESCRIPTION #######################################################
########################################################################
##   Parses data on the events given in SOD .csv format and puts into ##
##  individually formatted files in a directory.                      ##
##                                                                    ##
##   LAST EDITED: Jun. 14, 2008 at 20:10 GMT by Garrett Euler         ##
##   VERSION INFO:                                                    ##
##            1.3: user can now set chans, time limits, and units     ##
##            1.2: extended window (-1 to 3 hours), 1 sps data        ##
##            1.1: cleaned up version                                 ##
##            1.0: initial version                                    ##
########################################################################

#### USAGE #############################################################
########################################################################
##  Command line input:                                               ##
##     cutcamdb.pl ev_file ev_out_dir out_data_dir db chans           ##
##                 begin_time end_time counts nproc                   ##
##                                                                    ##
##     where:                                                         ##
##        *ev_file - SOD .csv format event file                       ##
##        *ev_out_dir - output event files directory                  ##
##        *out_data_dir - output data directory for extracted data    ##
##        *db - local database to pull data from                      ##
##        *chans - channel code (eg BHZ_02 or \* or LH[ZEN]_01 ...)   ##
##        *start_time - start time of cut relative to event (sec)     ##
##        *end_time - end time of cut relative to event (sec)         ##
##        *counts - 1 = units in counts, 0 = ground units             ##
##        *nproc - number of processor threads for this task          ##
########################################################################

_END_
	print $help;
	exit(1);
}

# EVENT FILE, NEW EVENT FILES DIRECTORY, NEW DB DATA DIR, DATABASE, CHANS
# RELATIVE TIME START AND END, COUNTS OPTION, NUMBER OF PROCESSES
my ($efile,$edir,$ndir,$db,$chan,$rts,$rte,$copt,$nproc)=@ARGV;

# CHECK INPUT EVENT FILE EXISTS
if (! -e $efile) {print "EVENT FILE $efile NOT FOUND!\n" and exit(2);}

# PUTTING SOD CSV TABLE INTO AN ARRAY
my @info=open_table($efile,1);
@info=shuffle(@info);

# PARSING THE INFO
unless (defined eval {mkpath($edir,0,0775)}) {print "COULD NOT CREATE DIRECTORY: $edir!\n" and exit(3);}
unless (defined eval {mkpath($ndir,0,0775)}) {print "COULD NOT CREATE DIRECTORY: $ndir!\n" and exit(4);}

# ABSOLUTE PATHS TO DIRS
my ($aedir,$andir)=abs_path_list($edir,$ndir);

# COUNTS OR GROUND UNITS
my $counts='';
if($copt==1) {$counts='-counts';}

# CHANNELS
my ($chans,$chan08)=("*:$chan","CM08:$chan");

# FORK OUT
my $id=forker($nproc);

# ASSIGN PARSELS
@info=parsel($id,$nproc,@info);

# LOOP THROUGH EACH EVENT
for (my $i=0; $i<=$#info; $i++) {
	# PARSE LINE BY COMMAS
	my @parts=split(', ',$info[$i]);
	
	# FURTHER PARSING DATE AND TIME FIELDS
	my $date=substr($parts[0],0,10);
	my $time=substr($parts[0],11,12);
	my ($evyear,$evmonth,$evday)=split('-',$date);
	my ($evhour,$evmin,$evsec)=split('\:',$time);
	my $evmil=$evsec-int($evsec);
	my $epoch=timegm($evsec,$evmin,$evhour,$evday,$evmonth-1,$evyear-1900)+$evmil;
	
	# FIND JULIAN DAY
	my $evjday=sprintf('%003d',julday($evyear,$evmonth-1,$evday-1)+1);
	
	# LOCATION
	my $evla=sprintf('%-9s',$parts[1]);
	my $evlo=sprintf('%-9s',$parts[2]);
	my $evdp=sprintf('%-9s',int($parts[3]*10+0.05)/10); # NEAREST TENTH KILOMETER
	
	# MAGNITUDE
	my $magtype=$parts[6]; my ($mb,$ms,$mw)=('','','');
	if ($magtype eq 'MB') {$mb=sprintf('%-9s',$parts[5])}
	elsif ($magtype eq 'MS') {$ms=sprintf('%-9s',$parts[5])}
	elsif ($magtype eq 'MW') {$mw=sprintf('%-9s',$parts[5])}
	
	# FORMATTING DATE/TIME
	my $kevnm=$evyear.".$evjday".".$evhour".".$evmin".".$evsec";
	$evyear=sprintf('%-9s',$evyear);
	$evjday=sprintf('%-9s',$evjday);
	$evhour=sprintf('%-9s',$evhour);
	$evmin=sprintf('%-9s',$evmin);
	$evsec=sprintf('%-9s',$evsec);
	
	# FORMATTED OUTPUT
	(my $eo=<<"	_END_")=~s/^\s+//gm;
		$kevnm  <= event title
		$evla  <= latitude
		$evlo  <= longitude
		$evdp  <= depth
		$evyear  <= year
		$evjday  <= julian day
		$evhour  <= hour
		$evmin  <= minute
		$evsec  <= seconds
		$mb  <= body wave magnitude
		$mw  <= moment magnitude
		$ms  <= surface wave magnitude
	_END_

	# WRITE OUT
	open(OUT,">$aedir$kevnm.event") or print "COULD NOT CREATE FILE: $kevnm.event!\n" and mexit(5); 
	print OUT $eo; close(OUT);
	
	# START AND END TIME
	my $ts=fancytime($epoch+$rts);
	my $te=fancytime($epoch+$rte);
	
	# RUNNING MAIN DB2SAC
	print "db2sac -i $counts -sc $chans -ts $ts -te $te -w $kevnm $db $andir$kevnm\n";
	`db2sac -i $counts -sc $chans -ts $ts -te $te -w $kevnm $db $andir$kevnm`;
	
	# DEAL WITH CM08 (ONLY FOR CERTAIN TIME FRAME)
	if ($evyear==2006 and $evjday<185) {
		# CORRECTION FOR CM08 TIMING OFFSET AND DRIFT
		my $cm08corr;
		if ($epoch<1150732800) {$cm08corr=689-($epoch-1136073600)*0.0000012594763;}
		else {$cm08corr=799.55-($epoch-1136073600)*0.0000012594763;}
		
		# START AND END TIME
		my $ts08=fancytime($epoch+$rts-$cm08corr);
		my $te08=fancytime($epoch+$rte-$cm08corr);
		
		# REMOVE PREVIOUS FILE
		my ($andatedir)=abs_path_list($andir.$kevnm);
		my @files=get_dir_files($andatedir);
		@files=grep {/CM08/} @files;
		@files=map {$_=$andatedir.$_} @files;
		foreach (@files) {
			unlink($_) or print "COULD NOT DELETE FILE $_!\n" and mexit(6);
		}
		
		# RUN FIXED CMO8 DB2SAC
		print "db2sac -i $counts -sc $chan08 -ts $ts08 -te $te08 -w $kevnm $db $andir$kevnm\n";
		`db2sac -i $counts -sc $chan08 -ts $ts08 -te $te08 -w $kevnm $db $andir$kevnm`;
	}
}

# MULTI-EXIT
mexit(0);

sub fancytime {
	my ($sec,$min,$hr,$day,$mon,$yr,$wkday,$yrday,$isdst)=gmtime($_[0]);
	return sprintf('%4d:%03d:%02d:%02d:%02.3f',$yr+1900,$yrday+1,$hr,$min,$sec+($_[0]-int($_[0])));
}
