#!/usr/bin/perl -w
use subs;
use strict;
use File::Path;
use List::Util qw(shuffle);
use FindBin qw($RealDir);
use lib $RealDir;
use MISC qw(open_table);
use DATETIME;
use FORKER;

# HELP MESSAGE
if ($#ARGV!=2) {
	my $help=<<'_END_';

#### DESCRIPTION #######################################################
########################################################################
##   Parses data on the events given in SOD .csv format and puts into ##
##  individually formatted files in a directory.                      ##
##                                                                    ##
##   LAST EDITED: Nov. 19, 2007 at 11:10am by Garrett Euler           ##
##   VERSION INFO:                                                    ##
##            2.4: file exists check, output directory path creation, ##
##                 and threaded support added to optimize efficiency  ##
##            2.3: code cleaning                                      ##
##            2.2: changed argument list for flexibility              ##
##            2.1: fixed julday bug (now gmt not local)               ##
##            2.0: adapted ev.pl for SOD usage                        ##
##            1.2: uses perl modules for julday                       ##
##            1.1: added this helpful feature.                        ##
##            1.0: initial version                                    ##
########################################################################

#### USAGE #############################################################
########################################################################
##  Command line input:                                               ##
##     evsod.pl ev_file ev_out_dir nproc                              ##
##                                                                    ##
##     where:                                                         ##
##        *ev_file - SOD .csv format event file                       ##
##        *ev_out_dir - output event files directory                  ##
##        *nproc - number of processor threads for this task          ##
########################################################################

_END_
	print $help;
	exit(1);
}

# EVENT FILE, NEW EVENT FILES DIRECTORY, NUMBER OF PROCESSES
my ($efile,$edir,$nproc)=@ARGV;

# CHECK INPUT EVENT FILE EXISTS
if (! -e $efile) {print "EVENT FILE $efile NOT FOUND!\n" and exit(2);}

# PUTTING SOD CSV TABLE INTO AN ARRAY
my @info=open_table($efile,1);
@info=shuffle(@info);

# PARSING THE INFO
unless (defined eval {mkpath($edir,0,0775)}) {print "COULD NOT CREATE DIRECTORY: $edir!\n" and exit(3);}

# MOVE INTO OUTPUT DIRECTORY
chdir($edir) or print "COULD NOT MOVE INTO OUTPUT DIRECTORY: $edir!\n" and exit(4);

# FORK OUT
my $id=forker($nproc);

# ASSIGN PARSELS
@info=parsel($id,$nproc,@info);

# LOOP THROUGH EACH EVENT
for (my $i=0; $i<=$#info; $i++) {
	# PARSE LINE BY COMMAS
	my @parts=split(', ',$info[$i]);
	
	# FURTHER PARSING DATE AND TIME FIELDS
	my $date=substr($parts[0],0,10);
	my $time=substr($parts[0],11,12);
	my ($evyear,$evmonth,$evday)=split('-',$date);
	my ($evhour,$evmin,$evsec)=split('\:',$time);
	
	# FINDING JULIAN DAY
	my $evjday=sprintf('%003d',julday($evyear,$evmonth-1,$evday-1)+1);
	
	# LOCATION
	my $evla=sprintf('%-9s',$parts[1]);
	my $evlo=sprintf('%-9s',$parts[2]);
	my $evdp=sprintf('%-9s',int($parts[3]*10+0.05)/10); # NEAREST TENTH KILOMETER
	
	# MAGNITUDE
	my $magtype=$parts[6]; my ($mb,$ms,$mw)=('','','');
	if ($magtype eq 'MB') {$mb=sprintf('%-9s',$parts[5])}
	elsif ($magtype eq 'MS') {$ms=sprintf('%-9s',$parts[5])}
	elsif ($magtype eq 'MW') {$mw=sprintf('%-9s',$parts[5])}
	
	# FORMATTING DATE/TIME
	my $kevnm=$evyear.".$evjday".".$evhour".".$evmin".".$evsec";
	$evyear=sprintf('%-9s',$evyear);
	$evjday=sprintf('%-9s',$evjday);
	$evhour=sprintf('%-9s',$evhour);
	$evmin=sprintf('%-9s',$evmin);
	$evsec=sprintf('%-9s',$evsec);
	
	# FORMATTED OUTPUT
	(my $eo=<<"	_END_")=~s/^\s+//gm;
		$kevnm  <= event title
		$evla  <= latitude
		$evlo  <= longitude
		$evdp  <= depth
		$evyear  <= year
		$evjday  <= julian day
		$evhour  <= hour
		$evmin  <= minute
		$evsec  <= seconds
		$mb  <= body wave magnitude
		$mw  <= moment magnitude
		$ms  <= surface wave magnitude
	_END_

	# WRITE OUT
	open(OUT,">$kevnm.event") or print "COULD NOT CREATE FILE: $kevnm.event!\n" and mexit(5); 
	print OUT $eo; close(OUT);
}

# MULTI-EXIT
mexit(0);


