#!/usr/bin/perl -w
use subs;
use strict;
use Fcntl qw(:DEFAULT :flock);
use File::Path;
use Time::Local;
use List::Util qw(shuffle);
use FindBin qw($RealDir);
use lib $RealDir;
use MISC qw(:other);
use DATETIME;
use FORKER;
use SAC;

# HELP MESSAGE
if ($#ARGV!=3) {
	my $help=<<'_END_';

#### DESCRIPTION #######################################################
########################################################################
##   Fills headers of sac data pulled from a database via db2sac.     ##
##  This version is for non-event specific data (like dayfiles as the ##
##  the name implies).                                                ##
##                                                                    ##
##  Files are renamed to a different sceme:                           ##
##                                                                    ##
##  			"FT.NW.STATN.ST.CMP.ID"                       ##
##   FT   -> filetype (UPPERCASE)                                     ##
##   NW   -> network code (1 to 2) (UPPERCASE)                        ##
##   STATN-> station name (1 to 5) (UPPERCASE)                        ##
##   ST   -> stream code (2) (integer)                                ##
##   CMP  -> component code (3) (UPPERCASE)                           ##
##   ID   -> multi id (2) (integer)                                   ##
##                                                                    ##
##                                                                    ##
##   LAST EDITED: Jun. 14, 2008 at 22:05pm by Garrett Euler           ##
##   VERSION INFO:                                                    ##
##            3.1: added fix for CM14, fix for channels with multiple ##
##                 files (added renaming of files basically)          ##
##            3.0: new version meant for dayfiles not event specific  ##
##            2.5: better avoidance of new file race conditions       ##
##            2.4: knetwk is an input now and some more header stuff  ##
##            2.3: lots of little bugfixes and added threading        ##
##            2.2: - removed timing file creation                     ##
##                 - perl modules added                               ##
##                 - attempts to be general for db2sac                ##
##            2.1: - fix for XB.CM02 for 2006.135 to 2007.060         ##
##                   (divide amplitude by 32)                         ##
##                 - fix for XB.CM08 for 2006.001 to 2006.184         ##
##                   (timeshift of 675+ or 798+ seconds)              ##
##            2.0: updated for new setup                              ##
##            1.2: better reflects timing as 'sac' sees it            ##
##            1.1: timing bug fixed                                   ##
##            1.0: initial version (replaced previous scripts)        ##
########################################################################

#### USAGE #############################################################
########################################################################
##  Command line input:                                               ##
##     fixdb2sac_dayfiles.pl in_data_dir out_data_dir knetwk nproc    ##
##                                                                    ##
##     where:                                                         ##
##        *in_data_dir - input data directory                         ##
##        *out_data_dir - output data directory                       ##
##        *knetwk - network code for data                             ##
##        *nproc - number of processor threads for this task          ##
########################################################################

_END_
	print $help;
	exit(1);
}

# INPUT DATA DIR, NEW DATA DIR, NETWORK NAME, NUMBER OF PROCESSES
my ($ddir,$ndir,$knetwk,$nproc)=@ARGV;

# CHECK INPUT DIRECTORY EXISTS
if (! -e $ddir) {print "DATA DIRECTORY $ddir NOT FOUND!\n" and exit(2);}

# CREATE OUTPUT DIRECTORY
unless (defined eval {mkpath($ndir,0,0775)}) {print "COULD NOT CREATE DIRECTORY: $ndir!\n" and exit(3);};

# RETRIEVING FILE LIST
my @files=get_dir_files($ddir);
@files=shuffle(@files);

# GETTING ABSOLUTE PATH TO NEW DATA DIRECTORY
my ($andir)=abs_path_list($ndir);

# MOVING INTO DATA DIRECTORY
chdir($ddir) or print "COULD NOT MOVE INTO DATA DIRECTORY: $ddir!\n" and exit(4);

# FORK OUT
my $id=forker($nproc);

# ASSIGN FILE PARSELS
@files=parsel($id,$nproc,@files);

# RETRIEVE MILLISECOND TIMING OF FILE
my @data=parselh(saclh(['nzmsec'],[@files]));

# LOOP THROUGH ALL FILES IN THE DIRECTORY
my (@station);
foreach (@data) {
	# GET FILENAME AND MILLISECS
	my ($file,$nzmsec)=@{$_};
	
	# REST OF TIMING FROM FILENAME (FASTER THAN SAC HEADER ACCESS)
	my ($datetime,undef,$kstnm,$cmploc)=split('\.',$file);
	my ($kcmpnm,$khole)=split('_',$cmploc);
	my $ft='SAC';
	my $nzyear=substr($datetime,0,4);
	my $nzjday=substr($datetime,4,3);
	my $nzhour=substr($datetime,7,2);
	my $nzmin=substr($datetime,9,2);
	my $nzsec=substr($datetime,11,2);
	
	# OUTPUT FILENAME (WITHOUT MULTI-SEGMENT ID)
	my $newname=$ft.".$knetwk".".$kstnm".".$khole".".$kcmpnm";
	
	# ENSURE OUTPUT FILE IS A NEW FILE
	my $repeat='00'; my $happy=0;
	until ($happy) {
		until (sysopen(NEWFILE,$andir.$newname.".$repeat",O_WRONLY|O_EXCL|O_CREAT)) {
			$repeat++; $repeat=sprintf('%02d',$repeat);
		}
		
		# MAKE SURE ONLY ONE PROCESS OWNS THIS FILE
		flock(NEWFILE,LOCK_EX) or print "COULD NOT LOCK FILE $newname.$repeat - RACE POSSIBLE!\n";
		
		# CHECK IF OUT RACED BY ANOTHER PROCESS OTHERWISE WRITE TO THE FILE 
		# (IF FILE HAS CONTENT NOW THIS PROCESS LOST THE RACE)
		open(LOCKEDNEWFILE,$andir.$newname.".$repeat"); my $dump=<LOCKEDNEWFILE>; close(LOCKEDNEWFILE);
		if (defined($dump)) {
			print "RACE AVOIDED!\n"; close(NEWFILE);
		} else {
			print NEWFILE $$; $happy=1; close(NEWFILE);
		}
	}
	
	# START TIME (FROM BEGINNING OF DAY)
	my $start=$nzhour*3600+$nzmin*60+$nzsec+$nzmsec/1000;
	
	# FIX CM08 TIMING
	if ($knetwk eq 'XB' and $kstnm eq 'CM08' and $nzyear==2006 and $nzjday<185) {
		# GET MONTH/DAY OF EVENT (FOR EPOCH TIME)
		my ($nzmonth,$nzday)=calday($nzyear,$nzjday-1);
		
		# FIX CM08 TIMING OFFSET AND DRIFT
		my $cm08corr; my $epoch=timegm($nzsec,$nzmin,$nzhour,$nzday+1,$nzmonth,$nzyear-1900)+$nzmsec/1000;
		
		# FIX CM08 TIMING OFFSET AND DRIFT
		###################################
		# OFFSET OF 689 OR 799.55 SECS (799.55 SEC OCCURS AFTER 1ST SERVICING)
		# DRIFTING AT ROUGHLY 1SEC/10DAYS
		# ACCURACY OF THIS CORRECTION IS PROBABLY/CONSERVATIVELY +/-0.5SECS
		# NOT CHANGING DELTA CAUSES SOME PRECISION LOSS
		# USING START OF RECORD TO GET CORRECTION
		# END OF RECORD MAY BE UP TO 0.1 SECS OFF (IN TERMS OF PRECISION AND IF 1 DAY RECORD)
		if ($epoch<1150732800) {$cm08corr=689-($epoch-1136073600)*0.0000012594763;}
		else {$cm08corr=799.55-($epoch-1136073600)*0.0000012594763;}
		$start+=$cm08corr;
	}
	
	# SIG DIG FIX
	$start=sprintf('%12.9g',$start);
	
	# SAC MACRO TO FILL/FIX HEADER
	(my $sac=<<"	_END_")=~s/^\s+//gm;
		production on
		r $file
		ch lovrok TRUE
		ch lcalda TRUE
		ch lpspol TRUE
		ch user7 -12345
		ch user8 -12345
		ch knetwk $knetwk
		ch khole $khole
		ch kcmpnm $kcmpnm
		ch kstnm $kstnm
		ch nzhour 0
		ch nzmin 0
		ch nzsec 0
		ch nzmsec 0
		ch b $start
	_END_

	# SAC MACRO TO FIX CM02 AMPLITUDES
	if ($knetwk eq 'XB' and $kstnm eq 'CM02' and (($nzyear==2006 and $nzjday>135) or ($nzyear==2007 and $nzjday<60))) {
		($sac.=<<"		_END_")=~s/^\s+//gm;
			div 32
		_END_
	}

	# SAC MACRO TO FIX CM14 EAST COMPONENT AMPLITUDES
	if ($knetwk eq 'XB' and $kstnm eq 'CM14' and ($kcmpnm eq 'LHE' or $kcmpnm eq 'BHE') and ($nzyear==2006 or $nzyear==2007)) {
		($sac.=<<"		_END_")=~s/^\s+//gm;
			mul 78.3
		_END_
	}

	# SAC MACRO FINISH
	($sac.=<<"	_END_")=~s/^\s+//gm;
		w dir $andir $newname.$repeat
		quit
	_END_

	# RUN SAC MACRO
	runsac($sac);
	
	# VERIFY
	if (! -e $andir.$newname.".$repeat") {print "COULD NOT FIX $file!\n" and mexit(6);}
}

# MULTI-EXIT
mexit(0);

