#!/usr/bin/perl -w
use subs;
use strict;
use File::Path;
use List::Util qw(shuffle);
use FindBin qw($RealDir);
use lib $RealDir;
use MISC qw(:other);
use FORKER;
use SAC;
use POSIX qw(fmod);

# HELP MESSAGE
if ($#ARGV<2 or $#ARGV>3) {
	my $help=<<'_END_';

#### DESCRIPTION #######################################################
########################################################################
##   Looks for oddities in orientation and fixes if minor. Also finds ##
##  orthogonal horizontal pairs and puts the perpendicular component  ##
##  orientation code in the kuser0 field. YOU SHOULD FOLLOW UP ON ANY ##
##  WARNINGS/MESSAGES OUTPUT BY THIS PROGRAM AND VERIFY THEY ARE THE  ##
##  CORRECT ACTION.  FAILURE TO DO SO WILL LIKELY LEAD TO PROBLEMS!!! ##
##                                                                    ##
##  Next Version: - deal with stations where stream code varies with  ##
##                  the component (rare but exists)                   ##
##                - report flip-flopped N and E channels or numbered  ##
##                - report unpaired horizontals                       ##
##                - report 3 or more horizontals                      ##
##                                                                    ##
##   LAST EDITED: Dec. 13, 2008 at 01:20 GMT by Garrett Euler         ##
##   VERSION INFO:                                                    ##
##            2.8: major fix (found out perl's modulus is int only)   ##
##            2.7: lots of fixes, still a crufty game                 ##
##            2.6: station specific fixes now in place                ##
##            2.5: sac macro sig dig fixes                            ##
##            2.4: better handling of when input and output           ##
##                 directories are equal (skip copying)               ##
##            2.3: threading support added to optimize efficiency     ##
##            2.2: found more orientation problems that need fixes    ##
##                 - fixes: Z,N,E specific fixes, orientation equal   ##
##                 but should be orthogonal fixes, and sac read fix   ##
##                 - added: can set minor angle now                   ##
##            2.1: code cleaning                                      ##
##            2.0: fixes minor orientation issues and finds           ##
##                 rotatable pairs                                    ##
##            1.1: no more tables - pull values from files and check  ##
##            1.0: initial version                                    ##
########################################################################

#### USAGE #############################################################
########################################################################
##  Command line input:                                               ##
##     fixorient.pl in_data_dir new_data_dir nproc [minor]            ##
##                                                                    ##
##     where:                                                         ##
##        *in_data_dir - input data directory                         ##
##        *new_data_dir - output data directory                       ##
##        *nproc - number of processor threads for this task          ##
##        *minor - angles less than this value will be deemed minor   ##
##                 and adjustments will be allowed - for instance if  ##
##                 a horizontal pair is almost orthogonal (92 deg     ##
##                 from one another) a value of 2 or more will allow  ##
##                 adjustment so they are orthogonal. In degrees.     ##
##                 Default is 3.                                      ##
########################################################################

_END_
	print $help;
	exit(1);
}

# DATA DIR, NEW DATA DIR, NUMBER OF PROCESSES
my ($ddir,$ndir,$nproc)=@ARGV;

# SET DEFAULTS
my $minor=3;
my $form='%12.9g';

# OPTIONAL ARGUMENTS
if ($#ARGV>=3) {$minor=$ARGV[3]} # MINOR CHANGE ANGLE

# CHECK DATA DIRECTORY EXISTS
if (!-e $ddir) {print "DATA DIRECTORY $ddir NOT FOUND!\n" and exit(2);}

# MAKING SURE OUTPUT DIRECTORY EXISTS
unless (defined eval {mkpath($ndir,0,0775)}) {print "COULD NOT CREATE DIRECTORY: $ndir!\n" and exit(3);}

# GETTING ABSOLUTE PATHS TO DIRECTORIES
my ($addir,$andir)=abs_path_list($ddir,$ndir);

# COPY FILES TO NEW DIRECTORY IF NEW DIRECTORY
if ($addir ne $andir) {
	print "COPYING DATA FILES IN $ddir TO $ndir...";
	my @files=get_dir_files($ddir);
	copy_list2dir($ndir,map {$_=$addir.$_} @files);
	print "FINISHED!\n";
}

# GET FILE LIST
my @files=get_dir_files($ndir);
@files=shuffle(@files);

# MOVING INTO DATA DIRECTORY
chdir($ndir) or print "COULD NOT MOVE INTO DATA DIRECTORY: $ndir!\n" and exit(4);

# READING IN CHANNEL INFO FROM DATAFILE HEADERS
my @data=parselh(saclh(['knetwk','kstnm','khole','kcmpnm','cmpinc','cmpaz'],[@files]));

# DISAMBIGUATE DATA STREAMS
# THIS FINDS WHICH CHANNELS ARE A SET
my %streams;
for (my $i=0; $i<=$#data; $i++) {
	# EXPAND RECORD
	my ($file,$knetwk,$kstnm,$khole,$kcmpnm)=@{$data[$i]};
	my ($src,$gc,$oc)=split('',$kcmpnm);
	
	# LOOK FOR NEW STREAM
	my $base="SAC.$knetwk.$kstnm.$khole.$src$gc";
	push(@{$streams{$base}},$i);
}
my @bases=keys %streams;

# FORK OUT
my $id=forker($nproc);

# ASSIGN PARSELS
@bases=parsel($id,$nproc,@bases);

# WORK THROUGH STREAMS
foreach my $base (@bases) {
	# WORK THROUGH STREAM FILES
	my (%mem,%inc,%az,%horz,%pr);
	foreach my $j (@{$streams{$base}}) {
		# EXPAND RECORD
		my ($file,$knetwk,$kstnm,$khole,$kcmpnm,$cmpinc,$cmpaz)=@{$data[$j]};
		my ($src,$gc,$oc)=split('',$kcmpnm);
		
		# SIG DIG FIX
		$cmpinc=sprintf($form,$cmpinc);
		$cmpaz=sprintf($form,$cmpaz);
		
		########################################################
		## BEGIN STATION SPECIFIC ORIENTATION FIXES
		########################################################
		# VERIFY THIS ISSUE WITH CZ.__.DPC (HOW LONG DOES THIS ISSUE LAST?)
		if ($knetwk eq 'CZ' and $kstnm eq 'DPC' and $khole eq '__' and $kcmpnm eq 'BHN' and $cmpaz==90) {
			print "$file IS KNOWN TO BE MISORIENTED!\n";
			print "==> ADJUSTING AZIMUTH TO 0\n";
			my $sac=chnhdr("r $file",'cmpaz',0,'w over');
			runsac($sac); $cmpaz=0;
		} elsif ($knetwk eq 'CZ' and $kstnm eq 'DPC' and $khole eq '__' and $kcmpnm eq 'BHE' and $cmpaz==180) {
			print "$file IS KNOWN TO BE MISORIENTED!\n";
			print "==> ADJUSTING AZIMUTH TO 90\n";
			my $sac=chnhdr("r $file",'cmpaz',90,'w over');
			runsac($sac); $cmpaz=90;
		}
		# IU.HNR HAS BHE = 175, BHN = 355, CURRENT SOLUTION => BHE => 265
		# POLARITY MIGHT BE WRONG
		if ($knetwk eq 'IU' and $kstnm eq 'HNR' and $kcmpnm eq 'BHE' and $cmpaz==175) {
			print "$file IS KNOWN TO MISORIENTED!\n";
			print "==> ADJUSTING AZIMUTH FROM 175 TO 265\n";
			my $sac=chnhdr("r $file",'cmpaz',265,'w over');
			runsac($sac); $cmpaz=265;
		}
		########################################################
		## END STATION SPECIFIC ORIENTATION FIXES
		########################################################
		
		# LOOKING OUT FOR NON HORIZONTAL/VERTICAL
		my $hflag=1;
		if (fmod($cmpinc,90)) {
			# PRINT MESSAGE
			print "$file IS SUPPOSEDLY NOT HORIZONTAL/VERTICAL\n";
			print "WITH AN INCLINATION OF $cmpinc\n";
			print "AND AN AZIMUTH OF $cmpaz!\n";
			
			# FIX IF MINOR (0,90,180 +/-MINOR ANGLE)
			my $carddir=int($cmpinc/90+0.5)*90;
			if (abs($cmpinc-$carddir)<=$minor) {
				print "INCLINATION IS MINOR - REMOVING\n\n";
				my $sac=chnhdr("r $file",'cmpinc',$carddir,'w over');
				runsac($sac); $cmpinc=$carddir;
			# MAJOR INCLINATION!
			} else {
				print "INCLINATION IS SIGNIFICANT - MANUAL HELP NEEDED IMMEDIATELY!\n";
				print "COULD THE CMPAZI and CMPINC BE SWITCHED???\n" and mexit(5);
			}
		}
		
		# EAST OR NORTH COMPONENT INCLINATION CHECKS (TRUSTING THE LABEL OVER THE NUMBER)
		if (($oc eq 'E' or $oc eq 'N') and ($cmpinc==0 or $cmpinc==180)) {
			# THIS IS BAD - SOMEBODY SCREWED UP THE
			# ORIENTATION(s) FOR THIS COMPONENT AND
			# PROBABLY THE OTHERS TOO
			print "$file IS LABELED AS A HORIZONTAL\n";
			print "BUT HAS AN INCLINATION OF $cmpinc\n";
			print "AND AN AZIMUTH OF $cmpaz!\n";
			print "==> ADJUSTING INCLINATION TO 90\n\n";
			my $sac=chnhdr("r $file",'cmpinc',90,'w over');
			runsac($sac); $cmpinc=90;
		}
		
		# VERTICAL (LABELED Z) INCLINATION CHECK
		if ($oc eq 'Z' and $cmpinc==90) {
			# MORE OF SAME AS MENTIONED IN THE LAST CHECK
			print "$file IS LABELED AS A VERTICAL\n";
			print "BUT HAS AN INCLINATION OF $cmpinc\n";
			print "AND AN AZIMUTH OF $cmpaz!\n";
			print "==> ADJUSTING INCLINATION TO 0\n\n";
			my $sac=chnhdr("r $file",'cmpinc',0,'w over');
			runsac($sac); $cmpinc=0;
		}
		
		# VERTICAL WITH NON-ZERO AZI CHECK (HELPFUL TO IDENTIFY ISSUES)
		# NOTE THAT THIS CAN ONLY BE DONE AFTER THE NAUGHTY
		# VERTICALLY INCLINED HORIZONTALS ARE FIXED - OTHERWISE
		# WE ARE ERASING THE AZIMUTH OF MISINCLINED HORIZONTALS
		if ($cmpinc==0 or $cmpinc==180) {
			# VERTICAL - BAD AZI?
			if ($cmpaz!=0) {
				print "$file IS SUPPOSEDLY A VERTICAL\n";
				print "WITH AN INCLINATION OF $cmpinc\n";
				print "BUT HAS AN AZIMUTH OF $cmpaz!\n";
				print "==> ADJUSTING AZIMUTH TO 0\n\n";
				my $sac=chnhdr("r $file",'cmpaz',0,'w over');
				runsac($sac); $cmpaz=0;
			}
			$hflag=0;
		}
		
		# CHECK AGAINST OTHER CHANNELS IN STREAM
		if (defined($mem{$oc})) {
			# LOOKING OUT FOR ORIENTATION CHANGE BETWEEN FILES FOR THE SAME CHANNEL
			if ($cmpaz ne $az{$oc} or $cmpinc ne $inc{$oc}) {
				print "$file orientation has changed!\n";
				print "INC: $inc{$oc} TO $cmpinc\n";
				print "AZI: $az{$oc} TO $cmpaz\n";
				
				# FIX IF MINOR (+/-MINOR ANGLE)
				if (abs($cmpaz-$az{$oc})<=$minor and abs($cmpinc-$inc{$oc})<=$minor) {
					print "DIFFERENCE IS MINOR - ADJUSTING\n";
					my $sac1=chnhdr("r $file",'cmpaz',$az{$oc},'w over');
					my $sac2=chnhdr("r $file",'cmpinc',$inc{$oc},'w over');
					runsac($sac1); runsac($sac2);
					$cmpaz=$az{$oc}; $cmpinc=$inc{$oc};
				} else {print "DIFFERENCE IS MAJOR - MANUAL HELP NEEDED IMMEDIATELY!\n" and mexit(6);}
			}
			
			# ADD MEMBER
			push(@{$mem{$oc}},$file);
		} else {
			# ADD CHANNEL
			$az{$oc}=$cmpaz;
			$inc{$oc}=$cmpinc;
			$mem{$oc}=[$file];
			$horz{$oc}=$hflag;
		}
	}
	
	# FIGURE OUT ORTHOGONAL HORIZONTAL PAIRS
	foreach my $oc1 (keys %az) {
		# CHECKING ONLY UNPAIRED HORIZONTALS
		if ($horz{$oc1}==1 and !defined($pr{$oc1})) {
			# SEARCH CHANNELS FOR A PERPENDICULAR
			foreach my $oc2 (keys %az) {
				# CHECKING ONLY UNPAIRED HORIZONTALS
				if ($horz{$oc2}==1 and $oc1 ne $oc2 and !defined($pr{$oc2})) {
					# MACRO READ STATEMENTS
					my ($read1,$read2)=('','');
					foreach (@{$mem{$oc1}}) {$read1.="r more $_\n";}
					foreach (@{$mem{$oc2}}) {$read2.="r more $_\n";}
					
					# CHECKING IF ORIENTED OPPOSITE OF EACH OTHER - FORCE TO
					# AZIMUTH IN RANGE -45 TO 135 AND THEN LET THEM GO THROUGH THE
					# NEXT SECTION
					if (abs($az{$oc1}-$az{$oc2})==180) {
						print "$base : ORIENTATIONS FOR $oc1 AND $oc2 ARE OPPOSITES!\n";
						print "AZIMUTHS ARE $az{$oc1} AND $az{$oc2}\n";
						if (($az{$oc1}<135 and $az{$oc1}>-45) or ($az{$oc1}>315 and $az{$oc1}<495)) {
							print "==> AZIMUTHS UNIFIED TO $az{$oc1}\n";
							my $sac=chnhdr($read2,'cmpaz',$az{$oc1},'w over');
							runsac($sac); $az{$oc2}=$az{$oc1};
						} else {
							print "==> AZIMUTHS UNIFIED TO $az{$oc2}\n";
							my $sac=chnhdr($read1,'cmpaz',$az{$oc2},'w over');
							runsac($sac); $az{$oc1}=$az{$oc2};
						}
					}
					
					# CHECK IF EQUAL - THIS HAPPENS OCCASIONALLY FROM SOMEONE
					# MESSING UP THE ORIENTATION INFORMATION (ANNOYED YET?)
					# FOR NORTH AND EAST COMPONENTS WE CAN SORT OUT REASONABLE 
					# SOLUTIONS - IF THEY ARE BOTH POINTED NORTH THEN THE EAST 
					# CHANNEL IS ORIENTED EAST - IF THEY ARE BOTH POINTED EAST
					# THEN THE NORTH CHANNEL IS ORIENTED NORTH - OTHERWISE THE
					# EAST COMPONENT IS ORIENTED 90 DEGREES CLOCKWISE FROM THE
					# AZIMUTHAL ORIENTATION THAT THEY SHARE AND A LOUD ERROR
					# MESSAGE IS PRESENTED EXPLAINING THAT WE SHOULD PRAY TO
					# THE GODS THAT THIS IS CORRECT.  THAT LAST SITUATION IS
					# HOW TWO NUMBERED HORIZONTALS ARE HANDLED - WITH THE GREATER 
					# NUMBERED CHANNEL BEING PUSHED TO 90 DEGREES CLOCKWISE.
					if ($az{$oc1}==$az{$oc2}) {
						print "$base : ORIENTATIONS FOR $oc1 AND $oc2 ARE EQUAL!\n";
						print "AZIMUTH IS $az{$oc1}\n";
						my $cw=sprintf($form,fmod($az{$oc1}+90,360));
						my $ccw=sprintf($form,fmod($az{$oc1}-90,360));
						if ($oc1 ne 'E' and $oc1 ne 'N') {
							# NUMBERED HORIZONTALS
							print "==> ORIENTING GREATER NUMBERED COMPONENT 90 DEG CLOCKWISE!\n";
							print "##########################################################\n";
							print "##########################################################\n";
							print "NOTE -- THIS IS TOTALLY A GUESS AND VERY LIKELY IS\n";
							print "        WRONG.  MOST NUMBERED COMPONENTS ARE ORDERED SO\n";
							print "        THAT THE GREATER IS 90 DEG CLOCKWISE FROM THE\n";
							print "        SMALLER - BUT THE FACT REMAINS THAT THESE WERE\n";
							print "        ORIENTED IN THE SAME DIRECTION.  ANOTHER POSSIBLE\n";
							print "        SOLUTION IS TO ROTATE THE SMALLER 90 DEG CCW.\n";
							print "##########################################################\n";
							print "##########################################################\n\n";
							if ($oc1<$oc2) {
								# PUSH SECOND
								my $sac=chnhdr($read2,'cmpaz',$cw,'w over');
								runsac($sac); $az{$oc2}=$cw;
							} else {
								# PUSH FIRST
								my $sac=chnhdr($read1,'cmpaz',$cw,'w over');
								runsac($sac); $az{$oc1}=$cw;
							}
						} elsif ($az{$oc1}==0) {
							print "==> ORIENTING EAST COMPONENT TO EAST!\n\n";
							if ($oc1 eq 'E') {
								my $sac=chnhdr($read1,'cmpaz',90,'w over');
								runsac($sac); $az{$oc1}=90;
							} else {
								my $sac=chnhdr($read2,'cmpaz',90,'w over');
								runsac($sac); $az{$oc2}=90;
							}
						} elsif ($az{$oc1}==90) {
							print "==> ORIENTING NORTH COMPONENT TO NORTH!\n\n";
							if ($oc1 eq 'E') {
								my $sac=chnhdr($read2,'cmpaz',0,'w over');
								runsac($sac); $az{$oc2}=0;
							} else {
								my $sac=chnhdr($read1,'cmpaz',0,'w over');
								runsac($sac); $az{$oc1}=0;
							}
						} elsif (($az{$oc1}<$minor and $az{$oc1}>-$minor) or ($az{$oc1}<180+$minor and $az{$oc1}>180-$minor) or ($az{$oc1}<360+$minor and $az{$oc1}>360-$minor)) {
							print "AZIMUTH IS NEARLY NORTH/SOUTH, SO REORIENTING EAST\n";
							print "COMPONENT WOULD SEEM THE BEST IDEA BUT THIS IS TOTALLY\n";
							print "A GUESS AND VERY LIKELY IS WRONG!!!\n";
							print "==> ORIENTING EAST COMPONENT 90 DEGREES CLOCKWISE!\n\n";
							if ($oc1 eq 'E') {
								my $sac=chnhdr($read1,'cmpaz',$cw,'w over');
								runsac($sac); $az{$oc1}=90;
							} else {
								my $sac=chnhdr($read2,'cmpaz',$cw,'w over');
								runsac($sac); $az{$oc2}=90;
							}
						} elsif (($az{$oc1}<90+$minor and $az{$oc1}>90-$minor) or ($az{$oc1}<270+$minor and $az{$oc1}>270-$minor)) {
							print "AZIMUTH IS NEARLY EAST/WEST, SO REORIENTING NORTH\n";
							print "COMPONENT WOULD SEEM THE BEST IDEA BUT THIS IS TOTALLY\n";
							print "A GUESS AND VERY LIKELY IS WRONG!!!\n";
							print "==> ORIENTING NORTH COMPONENT 90 DEGREES COUNTER-CLOCKWISE!\n\n";
							if ($oc1 eq 'E') {
								my $sac=chnhdr($read2,'cmpaz',$ccw,'w over');
								runsac($sac); $az{$oc2}=0;
							} else {
								my $sac=chnhdr($read1,'cmpaz',$ccw,'w over');
								runsac($sac); $az{$oc1}=0;
							}
						} else {
							# ODDLY ORIENTED EAST/NORTH HORIZONTALS
							print "==> ORIENTING EAST COMPONENT 90 DEG CLOCKWISE!\n";
							print "##########################################################\n";
							print "##########################################################\n";
							print "NOTE -- THIS IS TOTALLY A GUESS AND VERY LIKELY IS\n";
							print "        WRONG.  MOST EAST/NORTH COMPONENTS ARE ORIENTED\n";
							print "        SO THAT THE EAST CHANNEL IS 90 DEG CLOCKWISE FROM THE\n";
							print "        NORTH ONE - BUT THE FACT REMAINS HERE THAT THESE WERE\n";
							print "        ORIENTED NEITHER NORTH NOR EAST YET IN THE SAME\n";
							print "        DIRECTION.  ANOTHER JUST AS LIKELY SOLUTION IS TO\n";
							print "        ORIENT THE NORTH CHANNEL 90 DEG CCW.\n";
							print "##########################################################\n";
							print "##########################################################\n\n";
							if ($oc1 eq 'E') {
								my $sac=chnhdr($read1,'cmpaz',$cw,'w over');
								runsac($sac); $az{$oc1}=$cw;
							} else {
								my $sac=chnhdr($read2,'cmpaz',$cw,'w over');
								runsac($sac); $az{$oc2}=$cw;
							}
						}
					}
					
					# PERPENDICULARS
					my ($cw1,$ccw1)=(sprintf($form,fmod($az{$oc1}+90,360)),sprintf($form,fmod($az{$oc1}-90,360)));
					my ($cw2,$ccw2)=(sprintf($form,fmod($az{$oc2}+90,360)),sprintf($form,fmod($az{$oc2}-90,360)));
					
					# CHECK IF ALMOST PERPENDICULAR AND FIX IT
					if (($az{$oc2}-$cw1)>0 and ($az{$oc2}-$cw1)<=$minor) {
						# MINOR ADJUSTMENT TO ORIENTATION TO BE ORTHOGONAL
						print "$base : ORIENTATIONS $oc1 AND $oc2 ARE NEARLY ORTHOGONAL\n";
						print "==> ADJUSTING ORIENTATION OF $oc2 TO BE ORTHOGONAL\n\n";
						my $sac=chnhdr($read2,'cmpaz',$cw1,'w over');
						runsac($sac); $az{$oc2}=$cw1;
					} elsif (($az{$oc2}-$ccw1)>0 and ($az{$oc2}-$ccw1)<=$minor) {
						# MINOR ADJUSTMENT TO ORIENTATION TO BE ORTHOGONAL
						print "$base : ORIENTATIONS $oc1 AND $oc2 ARE NEARLY ORTHOGONAL\n";
						print "==> ADJUSTING ORIENTATION OF $oc2 TO BE ORTHOGONAL\n\n";
						my $sac=chnhdr($read2,'cmpaz',$ccw1,'w over');
						runsac($sac); $az{$oc2}=$ccw1;
					} elsif (($az{$oc1}-$cw2)>0 and ($az{$oc1}-$cw2)<=$minor) {
						# MINOR ADJUSTMENT TO ORIENTATION TO BE ORTHOGONAL
						print "$base : ORIENTATIONS $oc1 AND $oc2 ARE NEARLY ORTHOGONAL\n";
						print "==> ADJUSTING ORIENTATION OF $oc1 TO BE ORTHOGONAL\n\n";
						my $sac=chnhdr($read1,'cmpaz',$cw2,'w over');
						runsac($sac); $az{$oc1}=$cw2;
					} elsif (($az{$oc1}-$ccw2)>0 and ($az{$oc1}-$ccw2)<=$minor) {
						# MINOR ADJUSTMENT TO ORIENTATION TO BE ORTHOGONAL
						print "$base : ORIENTATIONS $oc1 AND $oc2 ARE NEARLY ORTHOGONAL\n";
						print "==> ADJUSTING ORIENTATION OF $oc1 TO BE ORTHOGONAL\n\n";
						my $sac=chnhdr($read1,'cmpaz',$ccw2,'w over');
						runsac($sac); $az{$oc2}=$ccw2;
					}
					
					# PERPENDICULARS (IN CASE THEY CHANGED)
					($cw1,$ccw1)=(sprintf($form,fmod($az{$oc1}+90,360)),sprintf($form,fmod($az{$oc1}-90,360)));
					($cw2,$ccw2)=(sprintf($form,fmod($az{$oc2}+90,360)),sprintf($form,fmod($az{$oc2}-90,360)));
					
					# CHECK IF PERPENDICULAR
					if ($az{$oc2}==$cw1 or $az{$oc2}==$ccw1 or $az{$oc1}==$cw2 or $az{$oc1}==$ccw2) {
						$pr{$oc1}=$oc2; $pr{$oc2}=$oc1;
						my $sac1=chnhdr($read1,'kuser0',$oc2,'w over');
						my $sac2=chnhdr($read2,'kuser0',$oc1,'w over');
						runsac($sac1); runsac($sac2); last;
					}
				}
			}
		}
	}
}

# MULTI-EXIT
mexit(0);

