#!/usr/bin/perl -w
use subs;
use strict;
use File::Path;
use List::Util qw(shuffle);
use FindBin qw($RealDir);
use lib $RealDir;
use MISC qw(:other);
use FORKER;
use SAC;

# HELP MESSAGE
if ($#ARGV!=2) {
	my $help=<<'_END_';

#### DESCRIPTION #######################################################
########################################################################
##   Fixes a few issues with SAC data acquired via SOD.               ##
##                                                                    ##
##   Current List of Fixes:                                           ##
##       *removes stream code from KCMPNM (redundant with KHOLE)      ##
##       *force KHOLE to have stream code '__' if stream is empty     ##
##       *changes SAC header value IZTYPE from IB (begin time) to IO  ##
##          (origin time) as this is the reference frame - this also  ##
##          fixes the end-time SOD bug (this is be deprecated for SOD ##
##          versions >2.2.2 but will not affect them adversely)       ##
##       *expects filenames to be of the format:                      ##
##            SAC.KNETWK.KSTNM.KHOLE.KCMPNM                           ##
##           or                                                       ##
##            SAC.KNETWK.KSTNM.KHOLE.KCMPNM.SEGMENT                   ##
##           output names take the second format.                     ##
##                                                                    ##
##   Note: Does not check if the new files will have naming           ##
##         collisions with preexisting files - will just overwrite.   ##
##         If there are multiple sources for files from an event then ##
##         unintended overwrites could happen.                        ##
##                                                                    ##
##   LAST EDITED: Nov. 24, 2008 at 23:40 GMT by Garrett Euler         ##
##   VERSION INFO:                                                    ##
##            2.5: change arguments - now give dir with event dirs    ##
##                 rather than the sod root directory                 ##
##            2.4: removed adding kevnm (not useful), added verbose   ##
##                 output for better logging                          ##
##            2.3: existence checks and threading                     ##
##                 removed issue with SAC's file limit                ##
##            2.2: fixes bugs added in last version                   ##
##            2.1: output directory added                             ##
##            2.0: error checks, keep '__' now                        ##
##            1.1: added in SAC header fixes                          ##
##            1.0: initial version                                    ##
########################################################################

#### USAGE #############################################################
########################################################################
##  Command line input:                                               ##
##     fixsod.pl old_data_dir new_data_dir nproc                      ##
##                                                                    ##
##     where:                                                         ##
##        *old_data_dir - input data directory of event dirs          ##
##        *new_data_dir - output data directory of event dirs         ##
##        *nproc - number of processing threads for this task         ##
########################################################################

_END_
	print $help;
	exit(1);
}

# OLD DATA DIR, NEW DATA DIR, NUMBER OF PROCESSES
my ($odir,$ndir,$nproc)=@ARGV;

# CHECK SOD DATA EXISTS
if (! -e $odir) {print "NO SOD DATA TO WORK WITH!\n" and exit(2);}

# SETUP OUTPUT DIRECTORY
unless (defined eval {mkpath($ndir,0,0775)}) {print "COULD NOT CREATE DIRECTORY: $ndir!\n" and exit(6);}

# GETTING ABSOLUTE PATHS TO DIRECTORIES
my ($andir,$aodir)=abs_path_list($ndir,$odir);

# GET DATE DIRECTORIES
my @dates=get_dir_dirs($odir);

# MOVE INTO DATA-SOD
chdir($odir) or print "COULD NOT MOVE INTO $odir DIRECTORY!\n" and exit(7);

# WORK THROUGH EACH DATE
foreach my $date (@dates) {
	# SKIP SPECIAL ENTRIES
	if ($date eq '.' or $date eq '..') {next;}
	
	# MESSAGE
	print "WORKING ON $date\n";
	
	# MAKE OUTPUT DATE DIRECTORY
	unless (defined eval {mkpath($andir.$date,0,0775)}) {print "COULD NOT CREATE DIRECTORY: $andir$date!\n" and exit(8);}
	
	# GETTING ABSOLUTE PATH TO OUTPUT DIRECTORY
	my ($andate)=abs_path_list($andir.$date);
	
	# GET DATAFILE LIST
	my @files=get_dir_files($date);
	@files=shuffle(@files);
	
	# MOVE INTO DATE DIRECTORY
	chdir($date) or print "COULD NOT MOVE INTO DATE DIRECTORY: $date!\n" and exit(9);
	print "NOW WORKING ON DATE: $date\n";
	
	# FORK OUT
	my $id=forker($nproc);
	
	# ASSIGN THREAD A PARSEL OF FILES
	@files=parsel($id,$nproc,@files);
	
	# WORK ON FILES INDIVIDUALLY
	foreach my $file (@files) {
		# SPLIT FILENAME BY PERIODS
		my @parts=split(/\./,$file);
		
		# SANITY CHECK
		if ($parts[0] ne 'SAC') {
			print "NON-STANDARD FILENAME: $file\n";
			print "FORMAT: SAC.NET_CODE.STN_NAME.STEAM_CODE.CHANNEL.SEGMENT\n";
			exit(10);
		}
		
		# NEW FILENAME (FORCE MULTISEGMENT FIELD)
		if (!defined($parts[5])) {$parts[5]='00';}
		else {$parts[5]=sprintf('%02d',$parts[5]);}
		my $newfile=join('.',@parts);
		
		# REMOVING STREAM CODE FROM KCMPNM OR ADDING THE STREAM CODE TO KHOLE
		my $sac;
		if ($parts[3] eq '__') {
			# FIXING KHOLE
			$sac=sacmacro("r $file","ch khole $parts[3]\nch iztype io","w dir $andate $newfile");
		} else {
			# FIXING KCMPNM
			$sac=sacmacro("r $file","ch kcmpnm $parts[4]\nch iztype io","w dir $andate $newfile");
		}
		
		# RUNNING SAC MACRO
		runsac($sac);
		
		# VERIFY
		if (! -e $andate.$newfile) {print "FAILED TO FIX $file FROM $date!\n" and mexit(11);}
		else {print "FIXED $file ==> RENAMED TO $newfile\n"}
	}
	
	# DETHREAD
	dethread(0,$id);
	
	# MOVE BACK INTO DATA-SOD
	chdir($aodir) or print "COULD NOT MOVE INTO $aodir DIRECTORY!\n" and exit(12);
}
exit(0);
