#!/usr/bin/perl -w
use subs;
use strict;
use File::Path;
use File::Copy;
use List::Util qw(shuffle first);
use FindBin qw($RealDir);
use lib $RealDir;
use MISC qw(:other);
use FORKER;
use SAC;

# HELP MESSAGE
if ($#ARGV!=2) {
	my $help=<<'_END_';

#### DESCRIPTION #######################################################
########################################################################
##   Interpolates time series data to a typical sample rate if it has ##
##  an atypical one.  Mainly for fixing a few stations that have had  ##
##  timing drift corrections applied and so the sample rate is off.   ##
##  Also fixes minor fluctuations in sample intervals that can cause  ##
##  problems later.                                                   ##
##                                                                    ##
##   LAST EDITED: Jun. 15, 2008 at 16:35 GMT by Garrett Euler         ##
##   VERSION INFO:                                                    ##
##            3.8: minor difference at max sig dig (8)                ##
##            3.7: updated xsacinterp                                 ##
##            3.6: handles when input and output directories are      ##
##                 equal better (skip copy when file unchanged)       ##
##            3.5: threading support added to optimize efficiency     ##
##            3.4: divided sample interval problems into 2 catagories ##
##                 - full out interpolation when the rate is          ##
##                   significantly different from a standard          ##
##                 - forced change of delta header value when there   ##
##                   is a minor difference from a standard            ##
##            3.3: back to fixing sample rate issues but with matlab  ##
##                 interpolation routines replacing the old sac ones  ##
##            3.2: now only prints a message for odd samplerates b/c  ##
##                 of sac interpolate having serious issues           ##
##            3.1: more standards added                               ##
##            3.0: perl modules and code cleaning                     ##
##            2.0: more exit codes and better interpolation           ##
##            1.2: removed close si option and added update           ##
##            1.1: more samplerates                                   ##
##            1.0: initial version                                    ##
########################################################################

#### USAGE #############################################################
########################################################################
##  Command line input:                                               ##
##     fixsr.pl in_data_dir new_data_dir nproc                        ##
##                                                                    ##
##     where:                                                         ##
##        *in_data_dir - input data directory                         ##
##        *new_data_dir - output data directory                       ##
##        *nproc - number of processor threads for this task          ##
########################################################################

_END_
	print $help;
	exit(1);
}

# DATA DIR, NEW DATA DIR, NUMBER OF PROCESSES
my ($ddir,$ndir,$nproc)=@ARGV;

# CREATE OUTPUT DIRECTORY
unless (defined eval {mkpath($ndir,0,0775)}) {print "COULD NOT CREATE DIRECTORY: $ndir!\n" and exit(2);};

# CHECK DATA DIRECTORY EXISTS
if (!-e $ddir) {print "DATA DIRECTORY $ddir NOT FOUND!\n" and exit(3);}

# GET DATAFILE LIST
my @files=get_dir_files($ddir);
@files=shuffle(@files);

# DEFINING STANDARD SAMPLE INTERVALS
my @stdint=qw(10 8 6 5 4 3 2 1
		0.875   0.8   0.75   0.625   0.6   0.5   0.4   0.375   0.32   0.3   0.25   0.24   0.2   0.16   0.125   0.1
		0.0875  0.08  0.075  0.0625  0.06  0.05  0.04  0.0375  0.032  0.03  0.025  0.024  0.02  0.016  0.0125  0.01
		0.00875 0.008 0.0075 0.00625 0.006 0.005 0.004 0.00375 0.0032 0.003 0.0025 0.0024 0.002 0.0016 0.00125 0.001);
my %linthash; @linthash{map {$_=sprintf('%5.5g',$_)} @stdint}=('')x@stdint;
my %hinthash; @hinthash{map {$_=sprintf('%8.8g',$_)} @stdint}=('')x@stdint;

# GETTING ABSOLUTE PATH TO OUTPUT DIRECTORY
my ($addir,$andir)=abs_path_list($ddir,$ndir);

# MOVING INTO DATA DIRECTORY
chdir($ddir) or print "COULD NOT MOVE INTO DATA DIRECTORY: $ddir!\n" and exit(4);

# FORK OUT
my $id=forker($nproc);

# ASSIGN FILE PARSELS
@files=parsel($id,$nproc,@files);

# GET HEADER INFO
my @data=parselh(saclh(['delta'],[@files]));

foreach (@data) {
	# EXPAND RECORD
	my ($file,$delta)=@{$_};
	my $ldelta=sprintf('%5.5g',$delta);
	my $hdelta=sprintf('%8.8g',$delta);
	
	# FIXING NONSTANDARD INTERVALS
	if (!defined($linthash{$ldelta})) {
		# NEW SAMPLE INTERVAL (~4 TIMES SMALLER TO PRESERVE WAVEFORMS)
		my $nsi=first {$_<=$delta/3.75} @stdint;
		if (!defined($nsi)) {$nsi=$stdint[$#stdint]} # FORCE HIGH SAMPLE RATES TO HIGHEST STANDARD
		$nsi=smash($nsi);
		
		# MESSAGE USER
		print "$file HAS A NONSTANDARD DELTA OF $delta ==> CHANGING TO $nsi\n";
		
		# RUN MATLAB SCRIPT
		`matlab -nosplash -nojvm -r xsacinterp\\(\\'$file\\'\\,\\'$andir$file\\'\\,1/$nsi\\)\\;exit\\;`;
	} elsif (!defined($hinthash{$hdelta})) {
		# ADMISSIBLE SAMPLE RATE DIFFERENCE - JUST ADJUST THE INTERVAL TO THE STANDARD
		my $sac=chnhdr("r $file",'delta',$ldelta,"w dir $andir $file");
		
		# RUN SAC MACRO
		print "$file HAS A MINOR DELTA DIFFERENCE ==> FORCED TO $ldelta\n";
		runsac($sac);
	} else {
		if ($andir ne $addir) {
			copy($file,$andir) or print "COULD NOT COPY $file TO $andir!\n" and mexit(6);
		}
	}
}

# WAIT TIL ALL CHILDREN FINISH
mexit(0);

