#!/usr/bin/perl -w
use subs;
use strict;
use Cwd qw(cwd);
use File::Path;
use File::Copy;
use FindBin qw($RealDir);
use lib $RealDir;
use MISC qw(:other);
use FORKER;


# HELP MESSAGE
if ($#ARGV!=2 and $#ARGV!=3) {
	my $help=<<'_END_';

#### DESCRIPTION #######################################################
########################################################################
##   Prepares seismic data for preprocessing.  Essentially this is a  ##
##  wrapper script for more specific scripts for streamlining and     ##
##  simplifying the pre-preprocessing stage.                          ##
##                                                                    ##
##   Stages:                                                          ##
##    <> immediate fixes dependent on data retrieval method           ##
##    <> setup extended directory structure                           ##
##    <> create event files                                           ##
##                                                                    ##
##   BUGS:                                                            ##
##    - currently fixsod and fixdb2sac do not check for filename      ##
##      collisions.  for now please ensure your datasets are unique   ##
##      or the fixed sod data will be lost (the original data is ok)  ##
##                                                                    ##
##   LAST EDITED: June 14, 2008 at 22:50 GMT by Garrett Euler         ##
##   VERSION INFO:                                                    ##
##            1.1: removed unneeded table dirs and fixed some bugs    ##
##            1.0: initial version                                    ##
########################################################################

#### USAGE #############################################################
########################################################################
##  Command line input:                                               ##
##     initialfix.pl db_dir fixed_dir nproc [knetwk]                  ##
##                                                                    ##
##     where:                                                         ##
##        *db_dir is top-level SOD directory for this dataset         ##
##        *fixed_dir - directory under the seismogram directory where ##
##                     acquired data that has undergone fixes goes    ##
##        *nproc - number of processes to use for tasks               ##
##        *knetwk - network name of the local database - if multiple  ##
##                  network names are needed then you will have to    ##
##                  make multiple runs of this script and move around ##
##                  directories to fool it. default is ZZ             ##      
########################################################################

_END_
	print $help;
	exit(1);
}

# ROOT DIR, FIXED DATA DIR, NUMBER OF PROCESSES
my ($rdir,$fixeddir,$nproc)=@ARGV;

# NETWORK NAME FOR LOCAL DATABASE DATA
my $knetwk='ZZ'; if ($#ARGV==3) {$knetwk=$ARGV[3]}

# MOVE INTO ROOT
chdir($rdir) or print "COULD NOT MOVE INTO DIRECTORY $rdir!\n" and exit(2);
my $ardir=cwd;

# MAKE DIRECTORY STRUCTURE
print "MAKING DIRECTORY STRUCTURE\n";
my ($logdir,$lstdir,$scrdir,$evdir,$sgdir,$pltdir,$docdir,$syndir,$cordir)=qw(logs lists scripts events seismograms plots docs synthetics corrections);
my @dirs=($logdir,$lstdir,$scrdir,$evdir,$sgdir,$pltdir,$docdir,$syndir,$cordir);
foreach (@dirs) {if (!-e $_) {mkdir($_) or print "COULD NOT MAKE DIRECTORY: $_!\n" and exit(3);}}

# ABSOLUTE DIRECTORY PATHS (WITH SEPARATOR ON THE END)
my ($alogdir,$alstdir,$ascrdir,$aevdir,$asgdir,$apltdir,$adocdir,$asyndir,$acordir)=abs_path_list(@dirs);

# MOVE FILES AROUND
print "MOVING SOD LOGS, LISTS AND SCRIPTS\n";
my @top=get_dir_files('.');
my @type=qw(.log .csv .xml);
foreach (my $i=0; $i<=$#type; $i++) {
	my @files=grep {/$type[$i]/} @top;
	foreach my $file (@files) {
		if (-f $file) {
			copy($file,$dirs[$i]) or print "COULD NOT COPY $file TO DIRECTORY $dirs[$i]!\n" and exit(4);
			unlink($file) or print "COULD NOT DELETE $file!\n" and exit(5);
		}
	}
}

# GENERATING EVENT INFO FILES FOR MINI-DB
print "MAKING EVENT FILES\n";
my @csv=grep {/csv/} get_dir_files($lstdir);
@csv=map {$_=$alstdir.$_} @csv;
foreach (@csv) {`evsod.pl $_ $evdir $nproc`;}

# FIXES FOR SOD
print "RUNNING SOD FIXES\n";
`fixsod.pl . $fixeddir $nproc`;

# FIXES FOR DB2SAC
chdir($sgdir) or print "COULD NOT MOVE INTO DIRECTORY $sgdir!\n" and exit(6);
if (-e 'data-db2sac' and -d 'data-db2sac') {
	print "RUNNING DB2SAC FIXES\n";
	unless (defined eval {mkpath($fixeddir,0,0775)}) {print "COULD NOT CREATE DIRECTORY: $fixeddir!\n" and exit(7);}
	my ($afixeddir)=abs_path_list($fixeddir);
	chdir('data-db2sac') or print "COULD NOT MOVE INTO DIRECTORY data-db2sac!\n" and exit(8);
	my @data_datetime=get_dir_dirs('.');
	foreach (@data_datetime) {
		`fixdb2sac.pl $_ $afixeddir$_ $aevdir$_.event XB $nproc`;
	}
}

