#!/usr/bin/perl -w
use subs;
use strict;

# HELP MESSAGE
if ($#ARGV!=7) {
	my $help=<<'_END_';

#### DESCRIPTION #######################################################
########################################################################
##   Further preparation of seismic data for analysis of a specific   ##
##  seismic phase.  Essentially this is a wrapper script for more     ##
##  specific scripts to simplify the preprocessing stage.             ##
##                                                                    ##
##   Stages:                                                          ##
##    <> pull data from indicated directory that has correct value of ##
##       the indicated header field (ie verticals - cmpinc=0)         ##
##    <> write phase arrival times into header                        ##
##    <> remove files which do not include window near phase time     ##
##    <> lowpass filter data to prepare for interpolation             ##
##    <> interpolate data to a low common sample rate to avoid edge   ##
##       effects caused by resampling without initial conditions      ##
##                                                                    ##
##   LAST EDITED: June 15, 2008 at 16:45 GMT by Garrett Euler         ##
##   VERSION INFO:                                                    ##
##            1.2: removed filtering and resampling                   ##
##            1.1: rewrite for new setup                              ##
##            1.0: initial version                                    ##
########################################################################

#### USAGE #############################################################
########################################################################
##  Command line input:                                               ##
##     investigate_phase.pl in_dir date field value phase 1dmodel     ##
##                          window nproc                              ##
##                                                                    ##
##     where:                                                         ##
##        *in_dir - input data directory (without date)               ##
##        *date - event datetime string of data to work on            ##
##        *field - header field to use for selecting files            ##
##        *value - value of header field for selecting files          ##
##        *phase - seismic phase of interest - case sensitive         ##
##        *1dmodel - 1D model used to get phase arrival time          ##
##        *window - begin and end times of a file must be further     ##
##                  in seconds than this value from the arrival time  ##
##                  to be kept (ie 1/2 of minimum allowed window)     ##
##        *nproc - number of processes to use for tasks               ##
########################################################################

_END_
	print $help;
	exit(1);
}

# INPUT DATA DIRECTORY, DATE TO WORK ON, HEADER FIELD, HEADER VALUE, SEISMIC
# PHASE, 1D REF MODEL, 1/2 MIN WINDOW, NUM OF PROC
my ($indir,$date,$field,$value,$phase,$model,$window,$nproc)=@ARGV;

# PULL DATA FROM DIRECTORY SELECTIVELY
open(WEED,">weed.in.$$"); print WEED "1\n2\n$field\n$value\n\n6\n"; close(WEED);
print "weed.pl $indir/$date $phase-maybe/$date < weed.in.$$ > ../logs/$date.$phase.weed.log; mv weed.in.$$ ../logs/$date.$phase.weed.in\n";
      `weed.pl $indir/$date $phase-maybe/$date < weed.in.$$ > ../logs/$date.$phase.weed.log; mv weed.in.$$ ../logs/$date.$phase.weed.in`;

# THIS IS THE POINT TO BAIL IF YOU HAVE MULTIPLE CRITERIA
print "CONTINUE? (Y/N)\n";
my $answer=<STDIN>;
chomp($answer); if(uc($answer) eq 'N') {print "User exited\n"; exit(0)}

# DETERMINE ARRIVAL TIMES AND PUT IN HEADER
print "arrival.pl $phase-maybe/$date $phase-arrival/$date $phase $model $nproc > ../logs/$date.$phase.arrival.log\n";
      `arrival.pl $phase-maybe/$date $phase-arrival/$date $phase $model $nproc > ../logs/$date.$phase.arrival.log`;

# REMOVE DATA WITHOUT ENOUGH TIME AROUND PHASE
open(WEED,">weed.in.$$"); print WEED "1\n5\ne t0\n-9999999 $window\n5\nt0 b\n-9999999 $window\n\n6\n"; close(WEED);
print "weed.pl $phase-arrival/$date $phase/$date < weed.in.$$ > ../logs/$date.$phase.weed.window.log; mv weed.in.$$ ../logs/$date.$phase.weed.window.in\n";
      `weed.pl $phase-arrival/$date $phase/$date < weed.in.$$ > ../logs/$date.$phase.weed.window.log; mv weed.in.$$ ../logs/$date.$phase.weed.window.in`;

