#!/usr/bin/perl -w
use subs;
use strict;
use File::Path;
use File::Copy;
use List::Util qw(shuffle);
use FindBin qw($RealDir);
use lib $RealDir;
use MISC qw(:other);
use FORKER;
use SAC;

# HELP MESSAGE
if ($#ARGV!=7 and $#ARGV!=8) {
	my $help=<<'_END_';

#### DESCRIPTION #######################################################
########################################################################
##   Finds segmented data and organizes overlaps, gaps, and multiples.##
##  If two segments are sequential (within a tolerance) they are      ##
##  merged.  Multiples and repeated segments are removed.  Large      ##
##  overlaps and gaps are noted (in STDOUT) and left alone.           ##
##                                                                    ##
##  Next Version: gap interpolation via matlab routines               ##
##                                                                    ##
##   LAST EDITED: Nov. 10, 2008 at  5:55 GMT by Garrett Euler         ##
##   VERSION INFO:                                                    ##
##            3.8: another workaround for SAC silent no-merge bug and ##
##                 verify is based on the samplerate                  ##
##            3.7: fixed tolerance bug (doesn't include an extra      ##
##                 sample interval now)                               ##
##            3.6: doc updates                                        ##
##            3.5: fixed delete bug in sub merge                      ##
##            3.4: better handling of when input and output           ##
##                 directories are equal (skip copying)               ##
##            3.3: - threading support added to optimize efficiency   ##
##                 - calling of matlab routines for gap interpolation ##
##                   established - just need routines                 ##
##            3.2: selection for overlaps or gaps only                ##
##            3.1: debugging and merge verification                   ##
##            3.0: major rewrite - highly modularized, filename       ##
##                 independence, some major bugfixes for complicated  ##
##                 situations and new options - gapfill, lapcut, ddr  ##
##                 and show                                           ##
##            2.4: removed nasty bug                                  ##
##            2.3: tableless                                          ##
##            2.2: perl modules and debugging                         ##
##            2.1: fix update bug                                     ##
##            2.0: cleaner, more portable, and resilient              ##
##            1.3: added update list output for tupdate.pl            ##
##            1.2: workaround for sac no-merge bug                    ##
##            1.1: streamlined and added this helpful feature.        ##
##            1.0: initial version                                    ##
########################################################################

#### USAGE #############################################################
########################################################################
##  Command line input:                                               ##
##     overgaps.pl in_data_dir out_data_dir tolerance type gapfill    ##
##        lapcut ddr nproc [show]                                     ##
##                                                                    ##
##     where:                                                         ##
##        *in_data_dir - input data directory                         ##
##        *out_data_dir - output data directory                       ##
##        *tolerance - merging tolerance (allowance in seconds for    ##
##          adjusting the timing of files)                            ##
##          -- remember the actual difference in timing between two   ##
##             mergible files is within the range:                    ##
##              sample_interval+/-tolerance                           ##
##          -- for records with high sample rates consider a value    ##
##             equivalent to a few sample intervals as SAC has single ##
##             precision timing which can be awful for merging long   ##
##             segments of what should be continuous data             ##
##          -- note that when merging multiple files the overall time ##
##             adjustment of any segment will never exceed this value ##
##          -- mergers adjust the timing of the shorter length file   ##
##             to make the files sequential                           ##
##          -- allowing greater than 1 second adjustment can clean up ##
##             leap second timing issues (in the immediate sense)     ##
##          -- negative values never allow merging                    ##
##        *type - 0 - work on both overlaps and gaps                  ##
##                1 - work on overlaps only                           ##
##                2 - work on gaps only                               ##
##        *gapfill - 0 - remove gaps by making files sequential       ##
##                   1 - fill gaps with zeros                         ##
##                   2 - interpolate gaps linearly                    ##
##                   3 - interpolate gaps piecewise cubic             ##
##                   4 - interpolate gaps with spline                 ##
##          -- only for gaps less than or equal to 'tolerance'        ##
##          -- interpolation options are not working yet!!!           ##
##        *lapcut - 0 - remove overlap by making files sequential     ##
##                  1 - cut overlap from shorter file and merge       ##
##          -- only for overlaps less than or equal to 'tolerance'    ##
##        *ddr - delta difference range in percent                    ##
##          -- ie. 5 means the sample interval of the two files must  ##
##             be within +/-5% of their averaged sample interval in   ##
##             order to merge or delete the files                     ##
##          -- 0 means the sample interval must be exactly the same   ##
##             for merge or delete to occur (RECOMMENDED!)            ##
##          -- note that files are not interpolated to a new sample   ##
##             rate - the shorter trace is adjusted to the rate of    ##
##             the longer - which will introduce distortion!          ##
##        *nproc - number of processor threads for this task          ##
##        *show - report unmerged overgaps (if defined)               ##
########################################################################

_END_
	print $help;
	exit(1);
}

# DATA DIR, NEW DATA DIR, MERGING TOLERANCE (SECONDS), SELECTION TYPE, GAP FILL TYPE, 
# OVERLAP CUT TYPE, DELTA DIFF RANGE (PERCENT), NUMBER OF PROCESSES, SHOW NONMERGING REPORTS
my ($ddir,$ndir,$tol,$type,$gapfill,$lapcut,$ddr,$nproc)=@ARGV[0..7];
my $show=0; if ($#ARGV==8) {$show=1}

# FIX DDR AND FORMATTING
$ddr/=100;
my $form='%12.9g';

# MAKING SURE OUTPUT DIRECTORY EXISTS
unless (defined eval {mkpath($ndir,0,0775)}) {print "COULD NOT CREATE DIRECTORY: $ndir!\n" and exit(3);}

# GETTING ABSOLUTE PATHS TO DIRECTORIES
my ($addir,$andir)=abs_path_list($ddir,$ndir);

# COPY FILES TO NEW DIRECTORY IF NEW DIRECTORY
if ($addir ne $andir) {
	print "COPYING DATA FILES IN $ddir TO $ndir...";
	my @files=get_dir_files($ddir);
	copy_list2dir($ndir,map {$_=$addir.$_} @files);
	print "FINISHED!\n";
}

# GET FILE LIST
my @files=get_dir_files($ndir);
@files=shuffle(@files);

# MOVING INTO DATA DIRECTORY
chdir($ndir) or print "COULD NOT MOVE INTO DATA DIRECTORY: $ndir!\n" and exit(3);

# GET HEADER INFO
my @data=parselh(saclh(['knetwk','kstnm','khole','kcmpnm','delta','b','e'],[@files]));

# DISAMBIGUATE CHANNELS
my %chans;
for (my $i=0; $i<=$#data; $i++) {
	# EXPAND RECORD
	my (undef,$knetwk,$kstnm,$khole,$kcmpnm)=@{$data[$i]};
	
	# INDEX CHANNEL FILES
	my $base="SAC.$knetwk.$kstnm.$khole.$kcmpnm";
	push(@{$chans{$base}},$i);
}
my @bases=keys %chans;

# FORK OUT
my $id=forker($nproc);

# ASSIGN PARSELS
@bases=parsel($id,$nproc,@bases);

# WORK THROUGH CHANNELS
foreach my $base (@bases) {
	# LOOP THROUGH EACH FILE REMOVING REPEATS
	looper(\@data,\@{$chans{$base}},$tol,$type,$gapfill,$lapcut,$ddr,$base,$show,$form,0);
	
	# LOOP THROUGH EACH FILE LOOKING FOR MERGES
	looper(\@data,\@{$chans{$base}},$tol,$type,$gapfill,$lapcut,$ddr,$base,$show,$form,1);
}
mexit(0);

sub looper {
	my ($data,$index,$tol,$type,$gapfill,$lapcut,$ddr,$base,$show,$form,$merge)=@_;
	
	# LOOPING THROUGH EACH FILE OF THIS CHANNEL
	for (my $j=0; $j<=$#{$index}; $j++) {
		# EXPAND RECORD
		my ($file1,undef,undef,undef,undef,$delta1,$b1,$e1)=@{$$data[$$index[$j]]};
		$delta1=sprintf($form,$delta1); $b1=sprintf($form,$b1); $e1=sprintf($form,$e1);
		
		# PROCEED ONLY IF NOT DELETED
		if (! -e $file1) {next;}
		
		# CHECK AGAINST OTHERS FOR MERGIBILITY
		for (my $k=$j+1; $k<=$#{$index}; $k++) {
			# EXPAND RECORD
			my ($file2,undef,undef,undef,undef,$delta2,$b2,$e2)=@{$$data[$$index[$k]]};
			$delta2=sprintf($form,$delta2); $b2=sprintf($form,$b2); $e2=sprintf($form,$e2);
			
			# PROCEED ONLY IF NOT DELETED AND NOT THE SAME FILE
			if (! -e $file2 or $file1 eq $file2) {next;}
			
			# CHECK SAMPLERATES ARE SOMEWHAT SIMILAR (+/-$ddr) OR SHORT CIRCUIT
			my $avgdelta=($delta1+$delta2)/2;
			if (abs($avgdelta-$delta1)/$avgdelta>$ddr or abs($avgdelta-$delta2)/$avgdelta>$ddr) {
				print "$file1 AND $file2 HAVE SIGNIFICANTLY DIFFERENT DELTA VALUES!\n";
				next;
			}
			
			# FIND MERGES OR REMOVE REPEATS
			if ($merge) {
				my $action;
				$action=order($file1,$delta1,$b1,$e1,$file2,$delta2,$b2,$e2,$tol,$type,$gapfill,$lapcut,$base,$show,$form);
				if ($action) {
					# JUMP IF FILE1 DELETED
					if (! -e $file1) {last;}
					
					# MERGED - FILE1 CHANGED, FILE2 DELETED
					# NOW UPDATING FILE1 ATTRIBUTES
					(undef,$b1,$e1,$delta1)=@{(parselh(saclh(['b','e','delta'],[$file1])))[0]};
					$delta1=sprintf($form,$delta1); $b1=sprintf($form,$b1); $e1=sprintf($form,$e1);
					
					# FORCE INNER LOOP TO RESET
					$k=$j; # LOOP WILL INCREMENT THIS TO $j+1
				}
			} else {piece($file1,$b1,$e1,$file2,$b2,$e2,$base)}
		}
	}
}

sub piece {
	my ($file1,$b1,$e1,$file2,$b2,$e2,$base)=@_;
	
	# CHECK FOR PARTIAL PIECES FIRST
	if ($b1==$b2) {if ($e1<$e2) {pp($file1,$file2,$base)} else {pp($file2,$file1,$base)}}
	elsif ($e1==$e2) {if ($b1<$b2) {pp($file2,$file1,$base)} else {pp($file1,$file2,$base)}}
	elsif ($b1<$b2 and $e1>$e2) {pp($file2,$file1,$base)}
	elsif ($b1>$b2 and $e1<$e2) {pp($file1,$file2,$base)}
}

sub pp {
	my ($file1,$file2,$base)=@_;
	print "$file1 IS A PARTIAL PIECE OF $file2 - DELETING $file1\n";
	unlink($file1) or print "COULD NOT DELETE $file1!\n" and mexit(10);
}

sub order {
	my ($file1,$delta1,$b1,$e1,$file2,$delta2,$b2,$e2,$tol,$type,$gapfill,$lapcut,$base,$show,$form)=@_;
	
	# DETERMINE ORDER
	my $first=0; if ($b1<$b2) {$first=1}
	
	# CHECK FOR TOLERABLE GAPS AND OVERLAPS
	if ($first) {return overgaps($file1,$delta1,$b1,$e1,$file2,$delta2,$b2,$e2,$tol,$type,$gapfill,$lapcut,$base,$show,$form,$first);}
	else {return overgaps($file2,$delta2,$b2,$e2,$file1,$delta1,$b1,$e1,$tol,$type,$gapfill,$lapcut,$base,$show,$form,$first);}
}

sub overgaps {
	my ($file1,$delta1,$b1,$e1,$file2,$delta2,$b2,$e2,$tol,$type,$gapfill,$lapcut,$base,$show,$form,$first)=@_;
	
	# LOOK OUT FOR PARTIAL PIECE
	if ($e1>$e2) {pp($file2,$file1,$base); return !$first;}
	
	# GAP OR OVERLAP
	my $delta=$delta1; if ($e1-$b1<$e2-$b2) {$delta=$delta2;}
	my $tdiff=$b2-$e1-$delta; my $atdiff=sprintf($form,abs($tdiff));
	my $gap=0; if ($tdiff>0) {$gap=1}
	
	# SHORT CIRCUIT FOR SELECTION TYPE
	if ($gap and $type==1) {return 0;}
	elsif (!$gap and $type==2) {return 0;}
	
	# MERGIBLE?
	if ($atdiff<=$tol) {
		# YES - ADJUST AND MERGE
		if ($gap) {print "MERGIBLE GAP: $file1 -- $file2 $atdiff SECONDS\n"}
		else {print "MERGIBLE OVERLAP: $file1 -- $file2 $atdiff SECONDS\n"}
		merge($file1,$delta1,$b1,$e1,$file2,$delta2,$b2,$e2,$gapfill,$lapcut,$first,$gap,$form);
		return 1;
	} else {
		# NO - OPTIONAL REPORT
		if ($show) {
			if ($gap) {print "LARGE GAP: $file1 -- $file2 $atdiff SECONDS\n"}
			else {print "LARGE OVERLAP: $file1 -- $file2 $atdiff SECONDS\n"}
		}
		return 0;
	}
}

sub merge {
	my ($file1,$delta1,$b1,$e1,$file2,$delta2,$b2,$e2,$gapfill,$lapcut,$first,$gap,$form)=@_;
	
	# DETERMINE SHORTER/LONGER 
	my ($shorter,$delta,$sfile)=(1,$delta2,$file1);
	if ($e1-$b1>$e2-$b2) {($shorter,$delta,$sfile)=(0,$delta1,$file2);}
	
	# FORCE DELTA CHANGE
	my $sac=chnhdr("r $sfile",'delta',$delta,'w over'); runsac($sac);
	
	# UPDATE SHORTER ATTRIBUTES
	($delta1,$b1,$e1,$delta2,$b2,$e2)=upatt($shorter,$file1,$delta1,$b1,$e1,$file2,$delta2,$b2,$e2,$form);
	
	# FIRST FILE
	my ($ffile,$nfile)=($file2,$file1); if ($first) {($ffile,$nfile)=($file1,$file2);}
	
	# TEMP FILE
	my $tfile="overgaps.merge.file.$$";
	
	# PLAN OF ACTION
	if ($gap and $gapfill==1) {
		# FILL GAP WITH ZEROS
		# NO ADJUSTMENT - JUST LET SAC DO ITS THING
	} elsif ($gap and $gapfill==2) {
		# LINEARLY INTERPOLATE GAP
		saclab_merge($file1,$file2,'linear',$ffile);
		unlink($nfile) or print "COULD NOT DELETE $nfile!\n" and mexit(20);
		verify($ffile,$b1,$e2,$delta);
		return;
	} elsif ($gap and $gapfill==3) {
		# PIECEWISE CUBIC INTERPOLATION OF GAP
		saclab_merge($file1,$file2,'pchip',$ffile);
		unlink($nfile) or print "COULD NOT DELETE $nfile!\n" and mexit(21);
		verify($ffile,$b1,$e2,$delta);
		return;
	} elsif ($gap and $gapfill==4) {
		# SPLINE INTERPOLATION OF GAP
		saclab_merge($file1,$file2,'spline',$ffile);
		unlink($nfile) or print "COULD NOT DELETE $nfile!\n" and mexit(22);
		verify($ffile,$b1,$e2,$delta);
		return;
	} elsif (!$gap and $lapcut==1) {
		# CUT OVERLAP FROM SHORTER
		if ($shorter) {$sac=chnhdr("cut z $b1 z $b2-$delta \nr $file1",'','','w over')}
		else {$sac=chnhdr("cut z $e1+$delta z $e2 \nr $file2",'','','w over')}
		runsac($sac);
		
		# UPDATE SHORTER ATTRIBUTES
		($delta1,$b1,$e1,$delta2,$b2,$e2)=upatt($shorter,$file1,$delta1,$b1,$e1,$file2,$delta2,$b2,$e2,$form);
		
		# FORCE TIMING OF SHORTER TO BE SEQUENTIAL
		my $shift=$e1+$delta; if ($shorter) {$shift=$b1+($b2-$e1)-$delta;}
		$sac=chnhdr("r $sfile",'b',sprintf($form,$shift),'w over'); runsac($sac);
		
		# UPDATE SHORTER ATTRIBUTES
		($delta1,$b1,$e1,$delta2,$b2,$e2)=upatt($shorter,$file1,$delta1,$b1,$e1,$file2,$delta2,$b2,$e2,$form);
		
		# MERGE
		$sac=chnhdr("r $file1\nmerge $file2",'','',"w $ffile");
		runsac($sac); unlink($nfile) or print "COULD NOT DELETE $nfile!\n" and mexit(23);
		verify($ffile,$b1,$e2,$delta);
	} else {
		# FORCE TIMING OF SHORTER TO BE SEQUENTIAL
		my $shift=$e1+$delta; if ($shorter) {$shift=$b1+($b2-$e1)-$delta;}
		$sac=chnhdr("r $sfile",'b',sprintf($form,$shift),'w over'); runsac($sac);
		
		# UPDATE SHORTER ATTRIBUTES
		($delta1,$b1,$e1,$delta2,$b2,$e2)=upatt($shorter,$file1,$delta1,$b1,$e1,$file2,$delta2,$b2,$e2,$form);
		
		# MERGE
		$sac=chnhdr("r $file1\nmerge $file2",'','',"w $tfile"); runsac($sac);
		
		# FUDGE FACTOR FIX IF SAC MERGE FAILS
		while (!verifytf($tfile,$b1,$e2,$delta)) {
			# FORCE TIMING OF SHORTER TO BE SEQUENTIAL WITH EXTRA PAD
			my $fudge=(rand 2)-1;
			$shift=$e1+$fudge*$delta; if ($shorter) {$shift=$b1+($b2-$e1)-$fudge*$delta;}
			$sac=chnhdr("r $sfile",'b',sprintf($form,$shift),'w over'); runsac($sac);
			
			# UPDATE SHORTER ATTRIBUTES
			($delta1,$b1,$e1,$delta2,$b2,$e2)=upatt($shorter,$file1,$delta1,$b1,$e1,$file2,$delta2,$b2,$e2,$form);
			
			# MERGE
			$sac=chnhdr("r $file1\nmerge $file2",'','',"w $tfile"); runsac($sac);
		}
		
		# CLEAN UP FILES
		copy($tfile,$ffile) or print "COULD NOT COPY $tfile TO $ffile!\n$!\n" and mexit(28);
		unlink($nfile) or print "COULD NOT DELETE $nfile!\n" and mexit(29);
	}
}

sub saclab_merge {
	print "THAR B UH BEG FALE HAR!\n" and mexit(30);
	#`matlab -nosplash -nojvm -r xsacmerge\\(\\'$_[0]\\'\\,\\'$_[1]\\'\\,\\'$_[2]\\'\\,\\'$_[3]\\'\\)\\;quit\\;`;
}

sub upatt {
	my ($shorter,$file1,$delta1,$b1,$e1,$file2,$delta2,$b2,$e2,$form)=@_;
	if ($shorter) {
		(undef,$b1,$e1,$delta1)=@{(parselh(saclh(['b','e','delta'],[$file1])))[0]};
		$delta1=sprintf($form,$delta1); $b1=sprintf($form,$b1); $e1=sprintf($form,$e1);
	} else {
		(undef,$b2,$e2,$delta2)=@{(parselh(saclh(['b','e','delta'],[$file2])))[0]};
		$delta2=sprintf($form,$delta2); $b2=sprintf($form,$b2); $e2=sprintf($form,$e2);
	}
	return ($delta1,$b1,$e1,$delta2,$b2,$e2);
}

sub verify {
	my ($file,$b1,$e2,$delta)=@_;
	my (undef,$b,$e)=@{(parselh(saclh(['b','e'],[$file])))[0]};
	if (abs($b-$b1)>$delta*0.5 or abs($e-$e2)>$delta*0.5) {
		print "MERGE FAILED FOR $file!\nEXPECTED BEGIN: $b1 -- ACTUAL BEGIN: $b\nEXPECTED END: $e2 -- ACTUAL END: $e\n" and mexit(40);}
}

sub verifytf {
	my ($file,$b1,$e2,$delta)=@_;
	my (undef,$b,$e)=@{(parselh(saclh(['b','e'],[$file])))[0]};
	if (abs($b-$b1)>$delta*0.5 or abs($e-$e2)>$delta*0.5) {return 0;} else {return 1;}
}

