#!/usr/bin/perl -w
use subs;
use strict;

# HELP MESSAGE
if ($#ARGV!=5) {
	my $help=<<'_END_';

#### DESCRIPTION #######################################################
########################################################################
##   Prepares seismic data for analysis.  Essentially this is a       ##
##  wrapper script for more specific scripts for streamlining and     ##
##  simplifying the preprocessing stage.                              ##
##                                                                    ##
##   Stages:                                                          ##
##    <> clean up orientation info                                    ##
##    <> clean up gaps/overlaps                                       ##
##    <> clean up samplerates                                         ##
##    <> toss out short data segments                                 ##
##    <> remove instrument mass drift                                 ##
##    <> remove instrument response                                   ##
##    <> rotate horizontals to radial/tranverse orientations          ##
##                                                                    ##
##   LAST EDITED: June 14, 2008 at 23:20 GMT by Garrett Euler         ##
##   VERSION INFO:                                                    ##
##            1.2: clean up                                           ##
##            1.1: rewrite for new setup                              ##
##            1.0: initial version                                    ##
########################################################################

#### USAGE #############################################################
########################################################################
##  Command line input:                                               ##
##     prepare_data.pl in_dir date short nproc ls lp                  ##
##                                                                    ##
##     where:                                                         ##
##        *in_dir - input data directory (without date)               ##
##        *date - event datetime string of data to work on            ##
##        *short - shortest datafile length allowed (in seconds)      ##
##        *nproc - number of processes to use for tasks               ##
##        *ls & lp - low freq taper for response removal              ##
########################################################################

_END_
	print $help;
	exit(1);
}

# INPUT DIRECTORY, DATE TO WORK ON, SHORT FILE LIMIT, NUMBER OF THREADS, LOW FREQ TAPER
my ($indir,$date,$short,$nproc,$ls,$lp)=@ARGV;

# CLEAN UP ORIENTATIONS
print "fixorient.pl $indir/$date data-fixorient/$date $nproc 3 > ../logs/$date.fixorient.log\n";
      `fixorient.pl $indir/$date data-fixorient/$date $nproc 3 > ../logs/$date.fixorient.log`;

# WAIT FOR INPUT
print "CONTINUE? (Y/N)\n";
my $answer=<STDIN>;
chomp($answer); if(uc($answer) eq 'N') {print "User exited\n"; exit(0)}

# CLEAN UP GAPS/OVERLAPS
print "overgaps.pl data-fixorient/$date data-overgaps/$date 2 0 0 0 10 $nproc > ../logs/$date.overgaps.log\n";
      `overgaps.pl data-fixorient/$date data-overgaps/$date 2 0 0 0 10 $nproc > ../logs/$date.overgaps.log`;

# CLEAN UP SAMPLERATES
print "fixsr.pl data-overgaps/$date data-fixsr/$date $nproc > ../logs/$date.fixsr.log\n";
      `fixsr.pl data-overgaps/$date data-fixsr/$date $nproc > ../logs/$date.fixsr.log`;

# REMOVE SHORT SEGMENTS AND DEAD CHANNELS
open(WEED,">weed.in.$$"); print WEED "1\n5\ne b\n0 $short\n\n6\n"; close(WEED);
print "weed.pl data-fixsr/$date data-weed/$date < weed.in.$$ > ../logs/$date.weed.log; mv weed.in.$$ ../logs/$date.weed.in\n";
      `weed.pl data-fixsr/$date data-weed/$date < weed.in.$$ > ../logs/$date.weed.log; mv weed.in.$$ ../logs/$date.weed.in`;

# CLEAN UP INSTRUMENT DRIFT
print "rdrift.pl data-weed/$date data-rdrift/$date $nproc 10 > ../logs/$date.rdrift.log\n";
      `rdrift.pl data-weed/$date data-rdrift/$date $nproc 10 > ../logs/$date.rdrift.log`;

# REMOVE INSTRUMENT RESPONSE
print "rinst.pl data-rdrift/$date data-rinst/$date pz ../pz data-rinstpz-fail/$date $nproc dis off $ls $lp > ../logs/$date.rinst.pz.log\n";
      `rinst.pl data-rdrift/$date data-rinst/$date pz ../pz data-rinstpz-fail/$date $nproc dis off $ls $lp > ../logs/$date.rinst.pz.log`;
print "rinst.pl data-rinstpz-fail/$date data-rinst/$date resp ../resp data-rinstresp-fail/$date $nproc dis off $ls $lp > ../logs/$date.rinst.resp.log\n";
      `rinst.pl data-rinstpz-fail/$date data-rinst/$date resp ../resp data-rinstresp-fail/$date $nproc dis off $ls $lp > ../logs/$date.rinst.resp.log`;

# ROTATE HORIZONTAL PAIRS
print "rotate.pl data-rinst/$date data-rotate/$date $short $nproc > ../logs/$date.rotate.log\n";
      `rotate.pl data-rinst/$date data-rotate/$date $short $nproc > ../logs/$date.rotate.log`;

