#!/usr/bin/perl -w
use subs;
use strict;
use File::Path;
use List::Util qw(shuffle);
use FindBin qw($RealDir);
use lib $RealDir;
use MISC qw(:other);
use FORKER;
use SAC;

# HELP MESSAGE
if ($#ARGV!=3) {
	my $help=<<'_END_';

#### DESCRIPTION #######################################################
########################################################################
##   Seismometers often deviate from a centered position as the       ##
##  instrumentation behaves inelastically, local temperatures         ##
##  fluctuate, underlying ground or pier settles, etc.  This can be   ##
##  remedied by detrending a seismogram (essentially performing a     ##
##  very-low corner highpass filter),  and finally tapering to remove ##
##  the potential for edge effects caused by truncation of the data   ##
##  stream.                                                           ##
##                                                                    ##
##  WARNING: The taper is not meant to be near phase arrivals as it   ##
##  modifies the frequency content.                                   ##
##                                                                    ##
##   LAST EDITED: Jun. 14, 2008 at 18:55 GMT by Garrett Euler         ##
##   VERSION INFO:                                                    ##
##            2.4: rtr; taper - simple again (removed rglitches)      ##
##            2.3: changed rglitches to just use default settings     ##
##            2.2: - rglitches added to clean up nasty                ##
##            2.1: - added ability to set grouping size as SAC has a  ##
##                 limit of 1000 files - smaller numbers limit impact ##
##                 of a failure while larger numbers are faster       ##
##                 - added verification that output files are made to ##
##                 idenitify problems rapidly                         ##
##                 - threading support added to optimize efficiency   ##
##            2.0: - error checks with exit codes                     ##
##                 - removed custom taper (not really that useful)    ##
##                 - no timing info necessary                         ##
##                 - assumes larger read limit of SAC 101.0           ##
##            1.2: - removed filtering (bad idea)                     ##
##                 - removed dependance on 'saclhdr' (slow)           ##
##                 - now uses timing list file,                       ##
##                 - added presorting and checks                      ##
##            1.1: - streamlined and added this help.                 ##
##            1.0: - initial version                                  ##
########################################################################

#### USAGE #############################################################
########################################################################
##  Command line input:                                               ##
##     rdrift.pl in_data_dir new_data_dir nproc group                 ##
##                                                                    ##
##     where:                                                         ##
##        *in_data_dir - input data directory                         ##
##        *new_data_dir - output data directory                       ##
##        *nproc - number of processor threads for this task          ##
##        *group - max number of files to process as a group, must be ##
##                  less than 1000                                    ##
########################################################################

_END_
	print $help;
	exit(1);
}

# DATA DIR, NEW DATA DIR, NUMBER OF PROCESSES, GROUP SIZE
my ($ddir,$ndir,$nproc,$group)=@ARGV;

# CHECK GROUP SIZE IS OK
if ($group>1000 or $group<1) {print "GROUP SIZE OUT OF RANGE" and exit(2);}

# CHECK INPUT DIRECTORY EXISTS
if (! -e $ddir) {print "DATA DIRECTORY $ddir NOT FOUND!\n" and exit(3);}

# CREATE OUTPUT DIRECTORY
unless (defined eval {mkpath($ndir,0,0775)}) {print "COULD NOT CREATE DIRECTORY: $ndir!\n" and exit(4);}

# GETTING ABSOLUTE PATHS TO DIRECTORIES
my ($andir)=abs_path_list($ndir);

# GET FILE LIST
my @files=get_dir_files($ddir);
@files=shuffle(@files);

# MOVING INTO DATA DIRECTORY
chdir($ddir) or print "COULD NOT MOVE INTO DATA DIRECTORY: $ddir!\n" and exit(5);

# FORK OUT
my $id=forker($nproc);

# ASSIGN FILE PARSELS
@files=parsel($id,$nproc,@files);

# LOOP THROUGH PARSELS AT SUBSET SIZE
for (my $i=0; $i<=$#files; $i+=$group) {
	# SET UPPER INDEX
	my $j=$i+$group-1; if ($j>$#files) {$j=$#files}
	
	# BUILD READ STATEMENT
	my $read=sacread(@files[$i..$j]);
	
	# BUILDING RDRIFT SAC MACRO
	my $sac=sacmacro($read,"rtrend\ntaper","w dir $andir over");
		
	# RUNNING SAC MACRO
	runsac($sac);
	
	# VERIFY
	foreach (@files[$i..$j]) {
		if (! -e $andir.$_) {print "FAILED TO DEDRIFT $_\n";}
	}
}

# WAIT TIL ALL CHILDREN FINISH
mexit(0);

