#!/usr/bin/perl -w
use subs;
use strict;
use Time::Local;
use File::Copy;
use File::Path;
use Cwd ('cwd');
use List::Util qw(shuffle);
use FindBin qw($RealDir);
use lib $RealDir;
use MISC qw(:other);
use DATETIME;
use FORKER;
use SAC;

# HELP MESSAGE
if ($#ARGV<5 or $#ARGV>11) {
	my $help=<<'_END_';

#### DESCRIPTION #######################################################
########################################################################
##   Seismometers are instruments that respond to earth motion. That  ##
##  response is unequal to the earth motion and is dependent on the   ##
##  amplitude and period of the earth's motion.  This is known as the ##
##  instrument response.                                              ##
##                                                                    ##
##  You will need to indicate which type of response file you are     ##
##  using (pz or resp only).  Requires 'sac' and maybe 'evalresp'.    ##
##                                                                    ##
##  Next Version: Get EVALRESP/FAPFILE option working - sniffing out  ##
##                a bug in sac (no error and no response removed???)  ##
##                                                                    ##
##   LAST EDITED: Dec. 12, 2008 at  5:10pm by Garrett Euler           ##
##   VERSION INFO:                                                    ##
##            2.7: handle KHOLE set to '__' when originally '' - note ##
##                 that we still cannot handle KHOLE set to ''        ##
##            2.6: split sac macro building to allow for including    ##
##                 station specific fixes                             ##
##            2.5: speedup for equal input and fail directories       ##
##            2.4: whitening option added                             ##
##            2.3: threading support added to optimize efficiency     ##
##            2.2: verify new file created check, adjustible ground   ##
##                 units and freqlimits (with sac defaults preset)    ##
##            2.1: no tables - accesses headers                       ##
##            2.0: complete rewrite - modularized and stuff           ##
##            1.4: 'production on' added to sac macros so SAC exits   ##
##                 on an error in instrument removal                  ##
##            1.3: new option 'fap' - uses external evalresp to make  ##
##                 freq-amp-phase files that are then put into sac    ##
##            1.2: deals with blank pz files                          ##
##            1.1: streamlined and added this help.                   ##
##            1.0: initial version                                    ##
########################################################################

#### USAGE #############################################################
########################################################################
##  Command line input:                                               ##
##     rinst.pl in_data_dir new_data_dir response_type response_dir   ##
##               fail_dir nproc [units] [whiten]                      ##
##               [low_stop] [low_pass] [high_pass] [high_stop]        ##
##     where:                                                         ##
##        *in_data_dir - input data directory                         ##
##        *new_data_dir - output data directory                       ##
##        *response_type - response file type used (pz,resp)          ##
##        *response_dir - path to the directory containing the        ##
##           response files                                           ##
##        *fail_dir - output data directory for failed responses      ##
##        *nproc - number of processor threads for this task          ##
##        *units - output ground units - acc, vel, dis (default dis)  ##
##        *whiten - prewhitening argument - on, off, or integer order ##
##           exactly like the sac 'whiten' options (default is off)   ##
##           -- integer order must be from 1 to 12                    ##
##        *low_stop, low_pass, high_pass, high_stop - frequency limit ##
##           corners that indicate stop and pass bands for applying   ##
##           the instrument response removal (defaults match SAC)     ##
########################################################################

_END_
	print $help;
	exit(1);
}

# DATA DIR, NEW DATA DIR, RESPONSE FILETYPE, RESPONSE DIR, FAIL DIR, NUMBER OF PROCESSES
my ($ddir,$ndir,$resp,$rdir,$fdir,$nproc)=@ARGV;

# SET DEFAULTS
my ($units,$whiten,$ls,$lp,$hp,$hs)=('none','off',-2,-1,1e5,1e6);

# READ IN OPTIONAL ARGUMENTS
if ($#ARGV>=6) {$units=$ARGV[6]}
if ($#ARGV>=7) {$whiten=$ARGV[7]}
if ($#ARGV>=8) {$ls=sprintf('%12.9g',$ARGV[8])}
if ($#ARGV>=9) {$lp=sprintf('%12.9g',$ARGV[9])}
if ($#ARGV>=10) {$hp=sprintf('%12.9g',$ARGV[10])}
if ($#ARGV>=11) {$hs=sprintf('%12.9g',$ARGV[11])}

# CHECK UNITS
$units=lc($units);
if ($units eq 'dis') {$units='none'}
if ($units ne 'none' and $units ne 'vel' and $units ne 'acc') {
	print "UNITS UNKNOWN - PLEASE CHOOSE ONE OF - ACC, VEL, OR DIS!\n" and exit(2);
}

# CHECK WHITEN
if ($whiten ne 'off' and $whiten ne 'on') {
	if ($whiten!=int($whiten) or $whiten<1 or $whiten>12) {print "WHITEN MUST BE AN INTEGER FROM 1 AND 12 OR ON OR OFF!\n" and exit(3);}
}

# CHECK INPUT DIRECTORY EXISTS
if (! -e $ddir) {print "DATA DIRECTORY $ddir NOT FOUND!\n" and exit(4);}

# CREATE OUTPUT DIRECTORY
unless (defined eval {mkpath($ndir,0,0755)}) {print "COULD NOT CREATE DIRECTORY: $ndir!\n" and exit(5);}
unless (defined eval {mkpath($fdir,0,0755)}) {print "COULD NOT CREATE DIRECTORY: $fdir!\n" and exit(6);}

# GETTING ABSOLUTE PATHS TO DIRECTORIES
my ($afdir,$andir,$addir,$ardir)=abs_path_list($fdir,$ndir,$ddir,$rdir);
my $cwd=cwd;

# CHECK IF FAIL DIRECTORY AND DATA DIRECTORY
my $failcopy=0; if ($afdir ne $addir) {$failcopy=1;}

# LIST OF INSTRUMENT RESPONSE FILES
my @rfiles=get_dir_files($ardir);

# GET LIST OF DATA FILES
my @files=get_dir_files($ddir);
@files=shuffle(@files);

# MOVING INTO DATA DIRECTORY
chdir($ddir) or print "COULD NOT MOVE INTO DATA DIRECTORY: $ddir!\n" and exit(7);

# FORK OUT
my $id=forker($nproc);

# ASSIGN FILE PARSELS
@files=parsel($id,$nproc,@files);

# GET HEADER INFO
my @data=parselh(saclh(['knetwk','kstnm','khole','kcmpnm','nzyear','nzjday','nzmonth','nzday','nzhour','nzmin','nzsec','nzmsec','b'],[@files]));

# PROCEED BY RESPONSE FILETYPE
if (lc($resp) eq 'resp') {
	# REMOVES INSTRUMENT RESPONSE USING SAC BUILT-IN EVALRESP
	foreach (@data) {
		my ($file,$knetwk,$kstnm,$khole,$kcmpnm)=@{$_};
		
		# FIND RESPONSE FILE
		my $respfile=$ardir."RESP.$knetwk.$kstnm.$khole.$kcmpnm";
		if (! -e $respfile and $khole eq '__') {
			$respfile=$ardir."RESP.$knetwk.$kstnm..$kcmpnm";
		}
		if (! -e $respfile) {
			print "\nRESPONSE FILE $respfile DOESN'T EXIST!\n";
			blown($file,$afdir,$failcopy); next;
		}
		
		# MACRO - READ AND RESPONSE REMOVAL
		(my $sac=<<"		_END_")=~s/^\s+//gm;
			production on
			r $file
			transfer from evalresp fname $respfile to $units freqlimits $ls $lp $hp $hs prewhitening $whiten
		_END_

		# MACRO - STATION SPECIFIC FIXES
		my @needdifresp=();
		my @needintresp=();
		foreach (@needdifresp) {if ($respfile eq $_) {$sac.="dif\n"; last;}}
		foreach (@needintresp) {if ($respfile eq $_) {$sac.="int\n"; last;}}
		
		# MACRO - WRITE AND EXIT
		$sac.="w dir $andir over\nquit";
		
		# RUNNING SAC MACRO AND CHECK AND HANDLE SAC FAILURE
		if (runsac2($sac)) {blown($file,$afdir,$failcopy); next;}
		
		# VERIFY NEW FILE EXISTS
		if (! -e $andir.$file) {blown($file,$afdir,$failcopy); next;}
		
		# ALL CLEAR
		print "$file RESPONSE REMOVAL SUCCESSFUL\n";
	}
} elsif (lc($resp) eq 'pz') {
	# REMOVES INSTRUMENT RESPONSE USING SAC POLEZEROS
	foreach (@data) {
		my ($file,$knetwk,$kstnm,$khole,$kcmpnm,$nzyear,$nzjday,$nzmonth,$nzday,$nzhour,$nzmin,$nzsec,$nzmsec,$b)=@{$_};
		
		# GET EPOCH TIME OF DATAFILE BEGIN
		my $data_epoch=timegm($nzsec,$nzmin,$nzhour,$nzday,$nzmonth-1,$nzyear-1900)+$nzmsec/1000+$b;
		
		# FIND POLEZERO FILE
		my $pzfile="SAC_PZs_${knetwk}_${kstnm}_${kcmpnm}_${khole}_";
		
		# NOW FIND POLEZEROS THAT MATCH
		my @pzfiles=grep {/$pzfile/} @rfiles;
		
		# FIX IN CASE NO MATCH
		if($#pzfiles==-1 and $khole eq '__') {
			# FIND POLEZERO FILE
			$pzfile="SAC_PZs_${knetwk}_${kstnm}_${kcmpnm}__";
			
			# NOW FIND POLEZEROS THAT MATCH
			@pzfiles=grep {/$pzfile/} @rfiles;
		}
		
		# FIGURE OUT WHICH MATCHES TO THE TIME OF THE DATA
		my $low; undef($pzfile);
		foreach my $pzf (@pzfiles) {
			# PULL OUT TIMING FROM PZFILE NAME (NEED TO FIGURE OUT IF END TIME IS ATTACHED)
			my $position1=rindex($pzf,'_',35)+1;
			my $position2=length($pzf);
			if ($position2>50) {$position2=rindex($pzf,'_');}
			my $length=$position2-$position1;
			my ($pzyear,$pzjday,$pzhour,$pzmin,$pzsec,$pztmsec)=split(/\./,substr($pzf,$position1,$length));
			my ($pzmonth,$pzday)=calday($pzyear,$pzjday-1);
			my $pzepoch=timegm($pzsec,$pzmin,$pzhour,$pzday+1,$pzmonth,$pzyear-1900)+$pztmsec/10000;
			
			# GET TIME DIFFERENCE (+ MEANS DATA LATER, - MEANS DATA EARLIER)
			my $dt=$data_epoch-$pzepoch;
			
			# CLOSE - SHOULD BE CHECKED OUT!
			if (abs($dt)<86400) {print "WARNING: $file INSTRUMENTATION INSTALLED/MODIFIED WITHIN 24 HOURS OF DATA!\n"}
			
			# USE MOST RECENT BEFORE EVENT
			if ($dt<0) {next;} # INSTRUMENTATION AFTER EVENT
			elsif (!defined($low)) {$low=$dt; $pzfile=$ardir.$pzf;}
			elsif ($dt<$low) {$low=$dt; $pzfile=$ardir.$pzf;}
		}
		
		# CHECK IF EXISTS
		if (!defined($pzfile) or !(-e $pzfile)) {
			print "\nPOLEZERO FILE CORRESPONDING TO TIMING OF $file DOES NOT EXIST!\n";
			blown($file,$afdir,$failcopy); next;
		}
			
		# CHECK IF EMPTY
		if(!open(PZ,$pzfile)) {print "COULD NOT OPEN POLEZERO FILE: $pzfile!\n"; blown($file,$afdir,$failcopy); next;}
		my @lines=<PZ>; close(PZ); chomp(@lines);
		if (join('',@lines) eq '') {print "POLEZERO FILE $pzfile IS EMPTY!\n"; blown($file,$afdir,$failcopy); next;}
		
		
		# MACRO - READ AND RESPONSE REMOVAL (MULTIPLIER TO CHANGE FROM METERS TO NANOMETERS)
		(my $sac=<<"		_END_")=~s/^\s+//gm;
			production on
			r $file
			transfer from polezero subtype $pzfile to $units freqlimits $ls $lp $hp $hs prewhitening $whiten
			mul 1e9
		_END_

		# MACRO - STATION SPECIFIC FIXES
		my @needdifpz=();
		my @needintpz=();
		foreach (@needdifpz) {if ($pzfile eq $_) {$sac.="dif\n"; last;}}
		foreach (@needintpz) {if ($pzfile eq $_) {$sac.="int\n"; last;}}
		
		# MACRO - WRITE AND EXIT
		$sac.="w dir $andir over\nquit";
		
		# RUNNING SAC MACRO AND CHECK AND HANDLE SAC FAILURE
		if (runsac2($sac)) {blown($file,$afdir,$failcopy); next;}
		
		# VERIFY NEW FILE EXISTS
		if (! -e $andir.$file) {
			# ADJUST CONSTANT
			# THIS CATCHES THE BUG IN SAC WHERE CERTAIN CONSTANT VALUES
			# WILL CAUSE TRANSFER TO CRASH (WITHOUT AN EXIT!)
			print "SAC FAILED TO APPLY POLEZEROS TO $file\n";
			print "CHECKING IF SAC CONSTANT BUG\n";
			open(PZ,">$pzfile");
			foreach my $line (@lines) {
				if ($line =~ /CONSTANT/) {
					my ($a,$b)=split(' ',$line);
					$line=join(' ',($a,sprintf('%12.6e',$b*1.00001)));
				}
				print PZ $line,"\n";
			}
			close(PZ);
			
			# RUNNING SAC MACRO AND CHECK AND HANDLE SAC FAILURE
			if (runsac2($sac) or !-e $andir.$file) {blown($file,$afdir,$failcopy); next;}
			
			# BUG FIXED
			print "INDEED IT IS ==> BUG FIXED\n"
		}
		
		# ALL CLEAR
		print "$file RESPONSE REMOVAL SUCCESSFUL\n";
	}
} elsif (lc($resp) eq 'fap') { print "OPTION DOES NOT WORK YET!\n" and mexit(10);
	# REMOVE INSTRUMENT RESPONSE USING EXTERNAL EVALRESP AND SAC
	foreach (@data) {
		my ($file,$knetwk,$kstnm,$khole,$kcmpnm,$nzyear,$nzjday)=@{$_};
		
		# FIND RESPONSE FILE
		my $respfile=$ardir."RESP.$knetwk.$kstnm.$khole.$kcmpnm";
		if (! -e $respfile) {
			print "\nRESPONSE FILE $respfile DOESN'T EXIST!\n";
			blown($file,$afdir,$failcopy); next;
		}
		
		# CALL EVALRESP
		`evalresp $kstnm $kcmpnm $nzyear $nzjday 0.0004 120 5000 -u dis -f $respfile`;
		
		# HANDLE EVALRESP CRASH
		my $exitcode = $? >> 8;  # exit code
		my $exitsign = $? & 127; # signal that caused exit
		if ($exitcode or $exitsign) {blown($file,$afdir,$failcopy); next;}
		
		# MAKE FAP FILE
		my (@temp,@phase); my $null=$khole; if ($khole eq '__') {$null='';}
		my $fapfile="$knetwk.$kstnm.$null.$kcmpnm";
		open(FAP,">FAP.$fapfile");
		@temp=open_table2("PHASE.$fapfile",0);
		for (my $i=0; $i<=$#temp; $i++) {(undef,$phase[$i])=split(" ",$temp[$i])}
		@temp=open_table2("AMP.$fapfile",0);
		print FAP ' theoretical  0   instrument    fap Organization',"\n";
		print FAP $#temp+1,"  \n";
		for (my $i=0; $i<=$#temp; $i++) {print FAP $temp[$i],'  ',$phase[$i],"  0.000000  0.000000 \n"}
		close(FAP);
		
		# BUILD RESPONSE REMOVAL SAC MACRO
		(my $sac=<<"		_END_")=~s/^\s+//gm;
			production on
			r $file
			transfer from fapfile subtype FAP.$fapfile to $units freqlimits $ls $lp $hp $hs prewhitening $whiten
			w dir $andir over
			quit
		_END_

		# RUNNING SAC MACRO AND CHECK AND HANDLE SAC FAILURE
		if (runsac2($sac)) {blown($file,$afdir,$failcopy); unlink("FAP.$fapfile"); unlink("AMP.$fapfile"); unlink("PHASE.$fapfile"); next;}
		
		# VERIFY NEW FILE EXISTS
		if (! -e $andir.$file) {blown($file,$afdir,$failcopy); next;}
		
		# REMOVE FAP FILES
		print "$file RESPONSE REMOVAL SUCCESSFUL\n";
		unlink("FAP.$fapfile"); unlink("AMP.$fapfile"); unlink("PHASE.$fapfile");
	}
} else {print "$resp IS NOT A SUPPORTED RESPONSE REMOVAL TYPE!\n" and mexit(8);}

# DETHREAD
dethread(0,$id);

# REMOVE MACRO FILES
@files=get_dir_files($addir);
@files=grep (/sac.in/,@files);
unlink_list(@files);

# HAPPY EXIT
exit(0);

sub blown {
	print "\nRESPONSE REMOVAL FAILED FOR $_[0]!\n";
	if ($_[2]) {
		print "MOVING TO CONTAINMENT DIRECTORY: $_[1]\n\n";
		copy($_[0],$_[1]) or print "COULD NOT COPY $_[0] TO $_[1]!\n" and mexit(101);
	}
}

sub open_table2 {
	open(TABLE,$_[0]) or print "COULD NOT OPEN TABLE FILE: $_[0]!\n" and exit(200);
	my @table=<TABLE>; close(TABLE); splice(@table,0,$_[1]); chomp(@table); return @table;
}
