#!/usr/bin/perl -w
use subs;
use strict;
use File::Path;
use List::Util qw(shuffle);
use Fcntl qw(:DEFAULT :flock);
use FindBin qw($RealDir);
use lib $RealDir;
use MISC qw(:other);
use FORKER;
use SAC;

if ($#ARGV!=3) {
	my $help=<<'_END_';

#### DESCRIPTION #######################################################
########################################################################
##   Finds all overlap between a pair of rotatable horizontals and    ##
##  makes rotated seismograms (orientation code is R or T).           ##
##                                                                    ##
##   Will make rotated components for any overlap greater than the    ##
##  specified amount of seconds.  Headers are updated.  Only rotated  ##
##  files are written to the output directory.                        ##
##                                                                    ##
##   LAST EDITED: Nov. 27, 2007 at  9:12pm by Garrett Euler           ##
##   VERSION INFO:                                                    ##
##            1.8: better avoidance of new file race conditions       ##
##            1.7: fixed fatal sac macro issue with too many sig dig  ##
##            1.6: output message and verification added              ##
##            1.5: threading support added to optimize efficiency     ##
##            1.4: no more tables - uses fixorient info put in kuser0 ##
##            1.3: output directory only contains rotated             ##
##            1.2: added in overlap cut to ensure stability           ##
##            1.1: streamlined and added this helpful feature.        ##
##            1.0: initial version                                    ##
########################################################################

#### USAGE #############################################################
########################################################################
##  Command line input:                                               ##
##     rotate.pl in_data_dir new_data_dir cutoff nproc                ##
##                                                                    ##
##     where:                                                         ##
##        *in_data_dir - input data directory                         ##
##        *new_data_dir - output data directory                       ##
##        *cutoff - minimum rotated file length desired (in seconds)  ##
##        *nproc - number of processor threads for this task          ##
########################################################################


_END_
	print $help;
	exit(1);
}

# DATA DIR, NEW DATA DIR, OVERLAP CUTOFF, NUMBER OF PROCESSES
my ($ddir,$ndir,$overlap,$nproc)= @ARGV;

# CHECK DATA DIRECTORY EXISTS
if (!-e $ddir) {print "DATA DIRECTORY $ddir NOT FOUND!\n" and exit(2);}

# MAKE OUTPUT DIRECTORY IF NEEDED
unless (defined eval {mkpath($ndir,0,0775)}) {print "COULD NOT CREATE DIRECTORY: $ndir!\n" and exit(3);};

# GET ABSOLUTE PATH TO OUTPUT DIRECTORY
my ($andir)=abs_path_list($ndir);

# GET FILE LIST
my @files=get_dir_files($ddir);
@files=shuffle(@files);

# MOVING INTO DATA DIRECTORY
chdir($ddir) or print "COULD NOT MOVE INTO DATA DIRECTORY: $ddir!\n" and exit(4);

# READING IN DATAFILE HEADER INFO
my @data=parselh(saclh(['knetwk','kstnm','khole','kcmpnm','kuser0','b','e','delta'],[@files]));

# DISAMBIGUATE HORIZONTAL PAIRS
my %pairs;
for (my $i=0; $i<=$#data; $i++) {
	# EXPAND RECORD
	my ($file,$knetwk,$kstnm,$khole,$kcmpnm,$oc2)=@{$data[$i]};
	my ($src,$gc,$oc1)=split('',$kcmpnm);
	
	# ONLY ALLOW PAIRED HORIZONTALS (FROM FIXORIENT)
	if (length($oc2)!=1) {next;}
	
	# LOOK FOR NEW PAIR
	my $base="SAC.$knetwk.$kstnm.$khole.$src$gc";
	if (defined($pairs{$base.$oc1.$oc2})) {push(@{${$pairs{$base.$oc1.$oc2}}[0]},$i);}
	elsif (defined($pairs{$base.$oc2.$oc1})) {push(@{${$pairs{$base.$oc2.$oc1}}[1]},$i);}
	else {push(@{${$pairs{$base.$oc1.$oc2}}[0]},$i);}
}
my @bases=keys %pairs;

# FORK OUT
my $id=forker($nproc);

# ASSIGN PARSELS
@bases=parsel($id,$nproc,@bases);

# WORK THROUGH PAIRS
foreach my $base (@bases) {
	# REMOVE ORIENTATION CODES
	my $sbase=substr($base,0,length($base)-2);
	
	# LOOKING FOR OVERLAPS - LOOP THROUGH EACH OF FIRST SET
	foreach my $j (@{${$pairs{$base}}[0]}) {
		my ($file1,undef,undef,undef,undef,undef,$b1,$e1,$d1)=@{$data[$j]};
		
		# CHECK FILE IS TOO SHORT
		if ($e1-$b1<$overlap) {next;}
		
		# CHECK AGAINST EACH OF SECOND SET
		foreach my $k (@{${$pairs{$base}}[1]}) {
			my ($file2,undef,undef,undef,undef,undef,$b2,$e2,$d2)=@{$data[$k]};
			
			# CHECK IF FILE IS TOO SHORT
			if ($e2-$b2<$overlap) {next;}
			
			# CHECK THAT SAMPLE RATES MATCH
			if ($d1!=$d2) {print "SAMPLE RATES OF $file1 and $file2 DO NOT MATCH!\n"; next;}
			
			# NOW FIGURE OUT POSITIONING
			if ($b1<=$b2) {
				# FILE1 BEGINS FIRST
				if ($e1-$b2>=$overlap) {
					# ENOUGH OVERLAP
					if ($e1<=$e2) {
						# FILE1 ENDS FIRST
						rotate($file1,$file2,$b2,$e1,$d1,$andir.$sbase);
					} else {
						# FILE2 ENDS FIRST
						rotate($file1,$file2,$b2,$e2,$d1,$andir.$sbase);
					}
				}
			} else {
				# FILE2 BEGINS FIRST
				if ($e2-$b1>=$overlap) {
					# ENOUGH OVERLAP
					if ($e1<=$e2) {
						# FILE1 ENDS FIRST
						rotate($file1,$file2,$b1,$e1,$d1,$andir.$sbase);
					} else {
						# FILE2 ENDS FIRST
						rotate($file1,$file2,$b1,$e2,$d1,$andir.$sbase);
					}
				}
			}
		}
	}
}

# MULTI-EXIT
mexit(0);

sub rotate {
	my ($file1,$file2,$b,$e,$dt,$base)=@_;
	
	# DISPLAY MESSAGE
	print "ROTATING $file1 AND $file2\n";
	
	# ENSURE OUTPUT FILE IS A NEW FILE
	my $rp='00'; my $happy=0;
	until ($happy) {
		until (sysopen(NEWFILE,$base."R.$rp",O_WRONLY|O_EXCL|O_CREAT)) {
			$rp++; $rp=sprintf('%02d',$rp);
		}
		
		# MAKE SURE ONLY ONE PROCESS OWNS THIS FILE
		flock(NEWFILE,LOCK_EX) or print "COULD NOT LOCK FILE ${base}R.$rp - RACE POSSIBLE!\n";
		
		# CHECK IF OUT RACED BY ANOTHER PROCESS OTHERWISE WRITE TO THE FILE 
		# (IF FILE HAS CONTENT NOW THIS PROCESS LOST THE RACE)
		open(LOCKEDNEWFILE,$base."R.$rp"); my $dump=<LOCKEDNEWFILE>; close(LOCKEDNEWFILE);
		if (defined($dump)) {
			print "RACE AVOIDED!\n"; close(NEWFILE);
		} else {
			print NEWFILE $$; $happy=1; close(NEWFILE);
		}
	}
	
	# CUT PARAMETERS
	my $ns=int(($e-$b)/$dt)+1;
	$b=sprintf('%12.9g',$b);
	
	# SAMPLE RATE AND GAIN CODE
	my $srg=substr($base,-2);
	
	# MAKING MACRO
	(my $sac=<<"	_END_")=~s/^\s+//gm;
		production on
		cut z $b n $ns
		r $file1 $file2
		rotate to gcp
		w ${base}R.$rp ${base}T.$rp
		r ${base}R.$rp
		chnhdr kcmpnm ${srg}R
		w over
		r ${base}T.$rp
		chnhdr kcmpnm ${srg}T
		w over
		quit
	_END_

	# RUNNING MACRO
	runsac($sac);
	
	# VERIFY
	if (! -e "${base}R.$rp") {print "ROTATION FAILED FOR $file1 AND $file2!\n$sac\n" and mexit(5)}
	if (! -e "${base}T.$rp") {print "ROTATION FAILED FOR $file1 AND $file2!\n$sac\n" and mexit(6)}
}

