#!/usr/bin/perl -w
use subs;
use strict;
use File::Path;
use FindBin qw($RealDir);
use lib $RealDir;
use MISC qw(:other);
use SAC;

# HELP MESSAGE
if ($#ARGV!=1) {
	my $help=<<'_END_';

#### DESCRIPTION #######################################################
########################################################################
##   Weeds out data files based upon header fields.  Input directory  ##
##  left intact unless output directory is the input directory.       ##
##                                                                    ##
##  Menu driven.  Will ask the user for input.                        ##
##                                                                    ##
##   LAST EDITED: Nov. 23, 2007 at  2:04pm by Garrett Euler           ##
##   VERSION INFO:                                                    ##
##            4.1: fixed bug where equal input and output directories ##
##                 doesn't remove weeded files                        ##
##            4.0: create your own weed/unweed                        ##
##            3.1: accesses header info on the fly                    ##
##            3.0: rewrite with modules                               ##
##            2.2: checks and cleaning                                ##
##            2.1: presorts lists so that unsorted lists are          ##
##                 supported as long as the entries are the same.     ##
##            2.0: bugfixes and fix for geometry format               ##
##            1.1: added arrival timing weed options and this         ##
##                 helpful feature.                                   ##
##            1.0: initial version                                    ##
########################################################################

#### USAGE #############################################################
########################################################################
##  Command line input:                                               ##
##     weed.pl in_data_dir new_data_dir                               ##
##                                                                    ##
##     where:                                                         ##
##        *in_data_dir - input data directory                         ##
##        *new_data_dir - output directory for the weeded dataset     ##
########################################################################

_END_
	print $help;
	exit(1);
}

# DATA DIR, NEW DATA DIR
my ($ddir,$ndir)=@ARGV;

# CHECK INPUT DIRECTORY EXISTANCE
if (!-e $ddir) {print "DATA DIRECTORY DOESN'T EXIST!\n" and exit(2);}

# MAKE SURE OUTPUT DIRECTORY EXISTS
unless (defined eval {mkpath($ndir,0,0775)}) {print "COULD NOT CREATE DIRECTORY: $ndir!\n" and exit(3);}

# GET ABSOLUTE PATH FOR DIRECTORIES
my ($addir,$andir)=abs_path_list($ddir,$ndir);

# FILE LIST
my @files=get_dir_files($ddir);
my $unweeded=[@files];
my ($fin,$fin2,$fin3);
my $weeded=[];

# MOVE INTO DATA DIRECTORY
chdir($ddir) or print "COULD NOT MOVE INTO DATA DIRECTORY: $ddir!\n" and exit(4);

# BEGIN USER MENU - LOOP UNTIL USER WANTS TO EXIT
until(defined($fin)) {
	# ALLOW ACCESS TO SUBMENUS
	undef($fin2);
	
	# MAIN MENU
	print "\n#####################################################\n";
	print "## NOTE :: WILDCARDS ARE _NOT_ ALLOWED!!!          ##\n";
	print "#####################################################\n";
	print "## WHICH OPTION?                                   ##\n";
	print "#####################################################\n";
	print "##  1  <==  WEED                                   ##\n";
	print "##  2  <==  UNWEED                                 ##\n";
	print "##  3  <==  LIST WEEDED FILES                      ##\n";
	print "##  4  <==  LIST UNWEEDED FILES                    ##\n";
	print "##  5  <==  RESET WEED                             ##\n";
	print "##  6  <==  COMMIT WEED AND EXIT                   ##\n";
	print "##  7  <==  JUST EXIT                              ##\n";
	print "#####################################################\n\n";
	
	# READ OPTION FROM USER
	my $option=<STDIN>; chomp($option);
	
	# PROGRESS BY OPTION
	if ($option eq '1' or $option eq '2') {
		#######################
		# CHANNEL SECTION
		#######################
		until (defined($fin2)) {
			# WEED OPTIONS
			my ($type,$field);
			print "\n#####################################################\n";
			print "## WHICH OPTION?                                   ##\n";
			print "#####################################################\n";
			print "##  1  <==  REMOVE                                 ##\n";
			print "##  2  <==  ALLOW                                  ##\n";
			print "##  3  <==  REMOVE RANGE                           ##\n";
			print "##  4  <==  ALLOW RANGE                            ##\n";
			print "##  5  <==  REMOVE DIFFERENCE RANGE                ##\n";
			print "##  6  <==  ALLOW DIFFERENCE RANGE                 ##\n";
			print "##                                                 ##\n";
			print "##     ANY OTHER OPTION WILL GO TO MAIN MENU       ##\n";
			print "#####################################################\n\n";

			# READ OPTION FROM USER
			my $suboption=<STDIN>; chomp($suboption);
			
			# PROGRESS BY OPTION
			if ($suboption eq '1') {$type='removed';}
			elsif ($suboption eq '2') {$type='allowed';}
			elsif ($suboption eq '3') {$type='out_rng';}
			elsif ($suboption eq '4') {$type='in_rng';}
			elsif ($suboption eq '5') {$type='outdrng';}
			elsif ($suboption eq '6') {$type='in_drng';}
			else {$fin2='yes';}
			
			if (defined($type)) {
				if ($type eq 'outdrng' or $type eq 'in_drng') {
					# HEADER VALUES
					print "\nWHICH HEADER VALUES WILL BE USED? (FIRST-SECOND) (SPACE SEPARATED)\n";
					
					# READ OPTION FROM USER
					$field=<STDIN>; chomp($field); $field=[split(' ',$field)];
					
				} else {
					# HEADER VALUE
					print "\nWHICH HEADER VALUE WILL BE USED?\n";
					
					# READ OPTION FROM USER
					$field=<STDIN>; chomp($field); ($field)=split(' ',$field);
				}
				
				# RUN WEED
				if (!defined($field) or ($suboption>4 and $#$field!=1)) {
					print "INVALID NUMBER OF FIELDS\n";
				} elsif ($option eq '1') {
					($unweeded,$weeded)=weed($unweeded,$weeded,$type,$field);
				} elsif ($option eq '2') {
					($weeded,$unweeded)=weed($weeded,$unweeded,$type,$field);
				}
				undef($type);
			}
		}
	} elsif ($option eq '3') {
		for (my $k=0; $k<=$#$weeded; $k++) {
			print (sprintf('%6d',$k+1),'  ',$$weeded[$k],"\n");
		}
	} elsif ($option eq '4') {
		for (my $k=0; $k<=$#$unweeded; $k++) {
			print (sprintf('%6d',$k+1),'  ',$$unweeded[$k],"\n");
		}
	} elsif ($option eq '5') {
		$unweeded=[@files];
		$weeded=[];
	} elsif ($option eq '6') {
		if ($addir eq $andir) {
			print "REMOVING WEEDED FROM $ddir...";
			unlink_list(@{$weeded});
			print "FINISHED!\n";
		} else {
			print "COPYING UNWEEDED TO $ndir...";
			copy_list2dir($andir,@{$unweeded});
			print "FINISHED!\n";
		}
		$fin='yes'; print "\n";
	} elsif ($option eq '7') {
		$fin='yes'; print "\n";
	} else {print "OPTION UNKNOWN ==> PLEASE TRY AGAIN.\n";}
}
exit(0);

sub weed {
	# INPUTS
	my ($unweeded,$weeded,$type,$field)=@_;
	
	# CHECK TYPE
	$type=uc($type); my $display;
	if ($type ne 'ALLOWED' and $type ne 'REMOVED'
	and $type ne 'IN_RNG' and $type ne 'OUT_RNG'
	and $type ne 'IN_DRNG' and $type ne 'OUTDRNG') {
		print "UNKNOWN WEED TYPE!\n" and exit(100);
	}
	
	# DIFFERENCE RANGE OPTION
	my @words;
	if ($type eq 'IN_DRNG' or $type eq 'OUTDRNG') {
		$display=uc($$field[0]).'-'.uc($$field[1]);
		print "\nDEFINE $display RANGE: (LOW HIGH) (SPACE SEPARATED)\n";
		my $line=<STDIN>; chomp($line); @words=split(' ',$line);
		if ($line eq '' or scalar(@words)!=2 or $words[0]>$words[1]) {
			print "Invalid Selection!\n";
			return $unweeded,$weeded;
		}
	# RANGE OPTION
	} elsif ($type eq 'IN_RNG' or $type eq 'OUT_RNG') {
		$display=uc($field); $field=[$field];
		print "\nDEFINE $display RANGE: (LOW HIGH) (SPACE SEPARATED)\n";
		my $line=<STDIN>; chomp($line); @words=split(' ',$line);
		if ($line eq '' or scalar(@words)!=2 or $words[0]>$words[1]) {
			print "Invalid Selection!\n";
			return $unweeded,$weeded;
		}
	# EXPLICIT
	} elsif ($type eq 'ALLOWED' or $type eq 'REMOVED') {
		$display=uc($field); $field=[$field];
		print "\nWHICH ${display}S DO YOU WANT ${type}? (SPACE SEPARATED)\n";
		my $line=<STDIN>; chomp($line); @words=split(' ',$line);
		if ($line eq '') {
			print "Invalid Selection!\n";
			return $unweeded,$weeded;
		}
	}
	
	# GRAB NETWORK CODES
	my @data=parselh(saclh($field,$unweeded));
	
	# CHECK THAT HEADER FIELD WAS VALID
	if (!@data) {
		print "NO FILES LEFT TO WEED OR INVALID HEADER FIELD!\n";
		return $unweeded,$weeded;
	}
	
	# TRACKING NUMBERS
	my ($n,$tn,$nbegin)=(0,0,scalar(@{$unweeded}));
	
	# TEMPORARY STORAGE
	my ($infiles,$outfiles,$indata,$outdata);
	
	# OPERATE BY TYPE
	if ($type eq 'IN_DRNG' or $type eq 'OUTDRNG') {
		@data=map {$_=[$$_[0],$$_[1]-$$_[2]]} @data;
		if ($type eq 'IN_DRNG') {
			($infiles,$outfiles)=range($words[0],$words[1],@data);
		} elsif ($type eq 'OUTDRNG') {
			($outfiles,$infiles)=range($words[0],$words[1],@data);
		}
		
		# ASSIGNMENT AND COUNT
		push(@{$weeded},@{$outfiles}); $unweeded=$infiles;
		($tn,$nbegin)=(scalar(@{$outfiles}),scalar(@{$infiles}));
		
	} elsif ($type eq 'IN_RNG' or $type eq 'OUT_RNG') {
		if ($type eq 'IN_RNG') {
			($infiles,$outfiles)=range($words[0],$words[1],@data);
		} elsif ($type eq 'OUT_RNG') {
			($outfiles,$infiles)=range($words[0],$words[1],@data);
		}
		
		# ASSIGNMENT AND COUNT
		push(@{$weeded},@{$outfiles}); $unweeded=$infiles;
		($tn,$nbegin)=(scalar(@{$outfiles}),scalar(@{$infiles}));
		
	} elsif ($type eq 'ALLOWED' or $type eq 'REMOVED') {
		# OUTPUT MESSAGE HEADER
		print " ___________________\n";
		print "|      RESULTS      |\n";
		print "|___________________|\n";
		print '| ',sprintf('%-7s',$display),' | ',sprintf('%7s',uc($type))," |\n";
		
		# PROCEED BY TYPE
		if ($type eq 'REMOVED') {
			# FIND BAD
			foreach my $bad (@words) {
				($outfiles,undef,undef,$indata)=parse($bad,@data);
				$n=scalar(@{$outfiles}); $tn+=$n; $nbegin-=$n;
				print '| ',sprintf('%-7s',$bad),' | ',sprintf('%7d',$n)," |\n";
				push(@{$weeded},@{$outfiles}); @data=@{$indata};
			}
			$unweeded=[map {($_)=@{$_}} @data];
		} elsif ($type eq 'ALLOWED') {
			# FIND GOOD
			$unweeded=[];
			foreach my $good (@words) {
				($infiles,undef,undef,$outdata)=parse($good,@data);
				$n=scalar(@{$infiles}); $tn+=$n; $nbegin-=$n;
				print '| ',sprintf('%-7s',$good),' | ',sprintf('%7d',$n)," |\n";
				push(@{$unweeded},@{$infiles}); @data=@{$outdata};
			}
			push(@{$weeded},map {($_)=@{$_}} @data);
			($tn,$nbegin)=($nbegin,$tn);
		}
		print "|_________|_________|\n";
	}
	
	# KEEPING TRACK OF NUMBERS
	print " ___________________\n";
	print "|       TOTAL       |\n";
	print "|___________________|\n";
	print "| ALLOWED | REMOVED |\n";
	print '| ',sprintf('%-7d',$nbegin),' | ',sprintf('%7d',$tn)," |\n";
	print "|_________|_________|\n";
	print "|      OVERALL      |\n";
	print "|___________________|\n";
	print "| ALLOWED | REMOVED |\n";
	print '| ',sprintf('%-7d',scalar(@{$unweeded})),' | ',sprintf('%7d',scalar(@{$weeded}))," |\n";
	print "|_________|_________|\n\n";
	
	return $unweeded,$weeded;
}

sub range {
	my ($low,$high,@record)=@_;
	my (@in,@out,@inrec,@outrec);
	foreach (@record) {
		my ($file,$value)=@{$_};
		if ($value>=$low and $value<=$high) {
			push(@in,$file);
			push(@inrec,[$file,$value]);
		} else {
			push(@out,$file);
			push(@outrec,[$file,$value]);
		}
	}
	return \(@in,@out,@inrec,@outrec);
}

sub parse {
	my ($good,@record)=@_;
	my (@in,@out,@inrec,@outrec);
	foreach (@record) {
		my ($file,$value)=@{$_};
		if ($value eq $good) {
			push(@in,$file);
			push(@inrec,[$file,$value]);
		} else {
			push(@out,$file);
			push(@outrec,[$file,$value]);
		}
	}
	return \(@in,@out,@inrec,@outrec);
}

